__version__ = '0.1.1'
__version_tuple__ = (0, 1, 1)


def get_version():
    from pathlib import Path
    error = ImportError

    if __version__ == '__UNSPECIFIED__':
        try:
            from versioningit import get_version, NotVersioningitError
            error = NotVersioningitError

            return get_version(Path(__file__).parent.parent)
        except error:
            return '999.999.999'
    else:
        return __version__


def get_version_tuple():
    from pathlib import Path
    error = ImportError

    if __version_tuple__ == '__UNSPECIFIED__':
        try:
            from versioningit import Versioningit, NotVersioningitError
            error = NotVersioningitError

            version_tuple_str = Versioningit\
                .from_project_dir(Path(__file__).parent.parent)\
                .do_template_fields(get_version(), None, None, None)['version_tuple']

            return eval(version_tuple_str)
        except error:
            return 999, 999, 999, '', ''
    else:
        return __version_tuple__



class MissingArgException(ValueError):
    def __init__(self, arg_name: str, arg_sources: list[str]):
        super().__init__(arg_name, arg_sources)

    @property
    def arg_name(self):
        return self.args[0]

    @property
    def arg_sources(self):
        return self.args[1]

    def __str__(self):
        arg_sources = [f'`{s}`' for s in self.arg_sources]
        ni = max(len(arg_sources) - 1, 1)
        suppliers = " or ".join([", ".join(arg_sources[:ni]), *arg_sources[ni:]])
        return f'Missing required arg `{self.arg_name}`,' \
               f' specify it with {suppliers}'


class MissingArgsException(Exception):
    def __init__(self):
        super().__init__([])

    @property
    def _missing_args(self) -> list[MissingArgException]:
        return self.args[0]

    def append(self, ex: MissingArgException):
        self._missing_args.append(ex)

    def __str__(self):
        return '\n'.join((str(arg) for arg in self._missing_args))



import threading
import time


class ValueObserver(threading.Thread):
    def __init__(self, value_supplier, callback, interval=0.5):
        threading.Thread.__init__(self)
        self.value_supplier = value_supplier
        self.callback = callback
        self.interval = interval
        self.last_n = 0
        self.running = True

        self.lock = threading.Lock()

    def sync(self, n):
        with self.lock:
            self.update(n - self.last_n)

    def update(self, dx):
        self.callback(dx)
        self.last_n += dx

    def run(self):
        while self.running:
            time.sleep(self.interval)
            with self.lock:
                n = self.value_supplier()
                dx = n - self.last_n
                self.update(dx)

    def stop(self):
        self.running = False



import numpy as np
from SimPEG.potential_fields.magnetics import Simulation3DIntegral
from discretize.utils import mkvc
from numpy.testing import assert_almost_equal

from metalpy.scab.builder import SimulationBuilder
from metalpy.scab.modelling import Scene
from metalpy.scab.modelling.shapes import Ellipsoid
from metalpy.scab.utils.misc import define_inducing_field
from metalpy.utils.sensor_array import get_grids_ex


magnetic_background_field = (50000, 78, 3)
magnetic_permeability = 2


def create_forward_settings():
    scene = Scene.of(
        Ellipsoid.spheroid(10, 30, 0),
    )
    model_mesh = scene.build(cell_size=1, cache=True)
    bounds = scene.bounds

    # observation
    obs = get_grids_ex(
        origin=[bounds.xmin, bounds.ymin, 50],
        end=[bounds.xmax, bounds.ymax, 50],
        n=[10 + 1, 10 + 1, 1]
    ).pts
    components = ['tmi']

    return model_mesh, obs, components


def create_simulation_builder(taichi=True):
    model_mesh, obs, components = create_forward_settings()

    builder = SimulationBuilder.of(Simulation3DIntegral)
    builder.receivers(obs, components)
    builder.active_mesh(model_mesh)
    builder.source_field(*magnetic_background_field)
    builder.model_type(scalar=True)

    if taichi:
        from metalpy.scab import Tied
        builder.patched(Tied(max_cpu_threads=1))

    return builder, model_mesh.model


def create_simpeg_simulation():
    from SimPEG import maps
    from SimPEG.potential_fields import magnetics

    source_field = define_inducing_field(*magnetic_background_field)
    model_mesh, receiver_points, components = create_forward_settings()

    nC = model_mesh.n_active_cells
    receiver_list = magnetics.receivers.Point(receiver_points, components=components)
    receiver_list = [receiver_list]

    inducing_field = magnetics.sources.SourceField(
        receiver_list=receiver_list, parameters=source_field
    )
    survey = magnetics.survey.Survey(inducing_field)
    model_map = maps.IdentityMap(nP=nC)
    simulation = magnetics.simulation.Simulation3DIntegral(
        survey=survey,
        mesh=model_mesh.mesh,
        model_type="scalar",
        chiMap=model_map,
        ind_active=model_mesh.active_cells,
        store_sensitivities="forward_only",
    )

    return simulation, model_mesh.model


def test_taichi_patch():
    # Taichi and Builder
    builder, model = create_simulation_builder(taichi=True)
    pred_taichi = builder.build().dpred(model * magnetic_permeability)

    # SimPEG
    simulation, model = create_simpeg_simulation()
    pred_simpeg = simulation.dpred(model * magnetic_permeability)

    # SimPEG新旧算法偏差最大能到千分之一 （Tied插件使用的是旧版本正演算法）
    assert_almost_equal(pred_taichi / pred_simpeg, 1, decimal=4)


def test_vector_model():
    field = define_inducing_field(*magnetic_background_field)

    # scalar model
    builder, model = create_simulation_builder(taichi=True)
    builder.model_type(scalar=True)
    pred_scalar = builder.build().dpred(model * magnetic_permeability)

    # vector model
    builder, model = create_simulation_builder(taichi=True)
    builder.model_type(vector=True)
    pred_vector = builder.build().dpred(mkvc(
        model[:, np.newaxis] * field.unit_vector * magnetic_permeability
    ))

    assert_almost_equal(pred_vector / pred_scalar, 1, decimal=7)



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, overload, Literal, Sequence
from xml import sax

import numpy as np
import pandas as pd

from metalpy.carto.coords import Coordinates
from metalpy.carto.coords.coordinates import CRSQuery
from metalpy.scab.modelling.shapes import Prism
from metalpy.scab.utils.proj import query_utm_crs_from_bound
from metalpy.utils.bounds import Bounds
from metalpy.utils.file import PathLike, make_cache_file_path, PathLikeType
from metalpy.utils.type import not_none_or_default

if TYPE_CHECKING:
    from metalpy.scab.modelling import Scene


class OSMHandler(sax.ContentHandler):
    def __init__(self):
        super().__init__()
        self.nodes = {}
        self.ways = {}
        self.buildings = {}
        self.heights = {}
        self.min_heights = {}
        self.current = None
        self.current_way = None
        self.bounds = np.zeros(4, dtype=np.float64)

    @property
    def utm_crs(self):
        return query_utm_crs_from_bound(self.bounds)

    def startElement(self, tag, attributes):
        if tag == 'bounds':
            self.bounds[0] = float(attributes['minlon'])
            self.bounds[1] = float(attributes['maxlon'])
            self.bounds[2] = float(attributes['minlat'])
            self.bounds[3] = float(attributes['maxlat'])
        elif tag == 'node':
            node_id = int(attributes['id'])
            self.nodes[node_id] = (float(attributes['lon']), float(attributes['lat']))
            self.current = tag
        elif tag == 'way':
            self.current_way = int(attributes['id'])
            self.current = tag
        elif tag == 'nd':
            if self.current == 'way':
                self.ways.setdefault(self.current_way, []).append(int(attributes['ref']))
        elif tag == 'tag':
            if self.current == 'way':
                if attributes['k'].startswith('building') and attributes['v'] != 'no':
                    # 既包含标准的building标签
                    # 也包含building:part标签，但可能会导致重叠
                    self.buildings[self.current_way] = self.ways[self.current_way]
                elif attributes['k'] == 'height':
                    self.heights[self.current_way] = float(attributes['v'])
                elif attributes['k'] == 'min_height':
                    self.min_heights[self.current_way] = float(attributes['v'])

    def endElement(self, tag):
        if tag in ("node", 'way'):
            self.current = None
            self.current_way = None


def load_osm(path: PathLike,
             level=0,
             default_height=10,
             height_map: dict[int, float] = None,
             dest_crs=Coordinates.SearchUTM
             ) -> tuple[list[Prism], OSMHandler]:
    """从OSM格式xml文件生成Scene实例

    Parameters
    ----------
    path
        OSM文件路径
    level
        目标区域的基准面海拔
    default_height
        OSM文件未指定高度的建筑物使用该默认高度替换
    height_map
        通过建筑物id，为建筑物指定高度
    dest_crs
        转换的目标坐标系（OSM文件中的建筑节点为WGS 84坐标系）

    Returns
    -------
    (scene, handler)
        给定OSM文件定义的建筑物列表和OSM文件相关信息

    Notes
    -----
    OSM格式坐标默认为经纬度，默认映射到UTM坐标系下，目前不支持跨UTM区
    """
    height_map = not_none_or_default(height_map, supplier=lambda: {})
    path = Path(path)

    parser = sax.make_parser()
    parser.setFeature(sax.handler.feature_namespaces, 0)
    handler = OSMHandler()
    parser.setContentHandler(handler)

    parser.parse(path)

    nodes = pd.DataFrame(handler.nodes).T
    Coordinates(nodes).warp(query=dest_crs, inplace=True)

    buildings = [
        Prism(nodes.loc[b[:-1]].to_numpy(),
              level + handler.min_heights.get(k, 0),
              level + handler.heights.get(k, height_map.get(k, default_height)))
        for k, b in handler.buildings.items()
    ]

    return buildings, handler


class OSMFormat:
    @staticmethod
    @overload
    def from_osm(
            path_or_bounds: PathLike | None = None,
            *,
            path: PathLike | None = None,
            level=0,
            default_height=10,
            height_map: dict[int, float] = None,
            extras: Literal[True],
            dest_crs: CRSQuery = Coordinates.SearchUTM
    ) -> tuple[Scene, OSMHandler]:
        ...

    @staticmethod
    @overload
    def from_osm(
            path_or_bounds: PathLike | None = None,
            *,
            path: PathLike | None = None,
            level=0,
            default_height=10,
            height_map: dict[int, float] = None,
            extras: Literal[False] = False,
            dest_crs: CRSQuery = Coordinates.SearchUTM
    ) -> Scene:
        ...

    @staticmethod
    @overload
    def from_osm(
            path_or_bounds: PathLike | None = None,
            *,
            path: PathLike | None = None,
            level=0,
            default_height=10,
            height_map: dict[int, float] = None,
            extras: bool,
            dest_crs: CRSQuery = Coordinates.SearchUTM
    ) -> tuple[Scene, OSMHandler] | Scene:
        ...

    @staticmethod
    def from_osm(
            path_or_bounds: PathLike | Sequence | None = None,
            *,
            path: PathLike | None = None,
            level=0,
            default_height=10,
            height_map: dict[int, float] = None,
            extras=False,
            dest_crs: CRSQuery = Coordinates.SearchUTM
    ) -> tuple[Scene, OSMHandler] | Scene:
        """从OSM格式xml文件生成Scene实例

        Parameters
        ----------
        path_or_bounds
            需要载入的OSM文件路径。
            或需要下载的建筑物区域WGS 84边界，
            格式依次为[minLong, maxLong, minLat, maxLat]
        path
            OSM文件缓存路径，只有
        level
            目标区域的基准面海拔
        default_height
            OSM文件未指定高度的建筑物使用该默认高度替换
        height_map
            通过建筑物id，为建筑物指定高度
        extras
            指示是否返回OSM解析中间信息
        dest_crs
            转换的目标坐标系（OSM文件中的建筑节点为WGS 84坐标系）

        Returns
        -------
        (scene, handler)
            给定OSM文件定义的建筑物列表和OSM文件相关信息

        Notes
        -----
        OSM格式坐标默认为经纬度，默认映射到UTM坐标系下，目前不支持跨UTM区

        OSM包含building和building:part两类建筑标签，
        其中building:part标签指定了建筑物的一部分，可能会和对应的building重叠。
        建议先在半透明模式下观察建筑分布情况，再视情况进行剔除，例如：

        >>> Scene.from_osm(bounds).to_multiblock().plot(opacity=0.5)
        """
        if isinstance(path_or_bounds, PathLikeType):
            path = Path(path_or_bounds)
        else:
            bounds = path_or_bounds

            bounds = Bounds(np.asarray(bounds)[:4])
            b_str = ','.join([str(i) for i in bounds.as_corners().ravel()])

            if path is None:
                path = make_cache_file_path(f'map{bounds}.osm')
            else:
                path = Path(path)

            if not path.exists():
                import requests
                url = f'https://overpass-api.de/api/map?bbox={b_str}'
                data = requests.get(url).content.decode('utf-8')
                with open(path, 'w', encoding='utf-8') as f:
                    f.write(data)

        assert path is not None, (
            'Either'
            ' a path to a local osm file'
            ' or bounds of a region to download osm data'
            ' must be specified.'
        )

        buildings, extra = load_osm(
            path=path,
            level=level,
            default_height=default_height,
            height_map=height_map,
            dest_crs=dest_crs
        )

        from metalpy.scab.modelling import Scene
        scene = Scene.of(*buildings)

        if extras:
            return scene, extra
        else:
            return scene



from __future__ import annotations

import warnings
from typing import Iterable, TYPE_CHECKING, Literal, overload

import numpy as np
from discretize import TensorMesh
from numpy.typing import ArrayLike

from metalpy.scab.modelling.object import Object
from metalpy.utils.numpy import ensure_as_numpy_array_collection
from metalpy.utils.type import is_numeric_array, get_first_key, ensure_set_key


if TYPE_CHECKING:
    import pyvista as pv


class ModelledMesh:
    """
    用于封装包含模型的网格。
    mesh是基础网格，active_cells是有效网格的选择器。
    其中总网格数为n_cells，active_cells选中的有效网格数为n_active_cells。

    model根据映射关系，区分为active_model和complete_model。
    其中:
    active_model长度为n_active_cells，每个元素直接对应一个有效网格的模型值。
    complete_model长度为n_cells，每个元素直接对应一个基础网格的模型值，其中无效网格的模型值为Scene.INACTIVE_{type}，
    type为模型的数据类型。

    Notes
    -----
    map_xxx_to_xxx系列函数均不保证输出的向量是否为原向量的视图，
    即输出的向量有可能与输入向量共享内存空间，也有可能不是，因此仅作只读用途使用。
    """

    def __init__(self,
                 mesh: TensorMesh,
                 models: dict[str, ArrayLike] | ArrayLike | None = None,
                 ind_active: ArrayLike = None,
                 **kwargs: ArrayLike):
        from metalpy.scab.modelling import Scene

        self._mesh = mesh

        default_key = None
        converted_models = ensure_as_numpy_array_collection(kwargs)

        if models is not None:
            if is_numeric_array(models):
                default_key = ensure_set_key(
                    converted_models,
                    Object.DEFAULT_KEY,
                    np.asarray(models)
                )
            else:
                converted_models.update(ensure_as_numpy_array_collection(models))

        if default_key is None:
            default_key = get_first_key(converted_models)

        if ind_active is not None:
            ind_active = np.asarray(ind_active)
            if ind_active.dtype == bool:
                assert ind_active.shape[0] == self.n_cells, \
                    f'`ind_active` defines activeness for all mesh cells,' \
                    f' which must have same size with mesh cells.' \
                    f' Got {ind_active.shape[0]}, expected {self.n_cells}.'
            else:
                assert np.issubdtype(ind_active.dtype, np.integer), \
                    f'`ind_active` must be array of `bool` or `integer`. Got `{ind_active.dtype}`.'
        else:
            if len(converted_models) > 0:
                for key, model in converted_models.items():
                    assert model.shape[0] == self.n_cells, \
                        f'`ModelledMesh`\'s constructor expects models to contain all mesh cells\'s values' \
                        f' in order to build `active_index`.' \
                        f' Avoid this by either specifying `active_index` manually or' \
                        f' converting models["{key}"] to match all the mesh cells.' \
                        f' Got {model.shape[0]}, expected {self.n_cells}.'
                ind_active = Scene.compute_active(converted_models.values())
            else:
                ind_active = np.ones(mesh.n_cells, dtype=bool)

        self._models = converted_models
        self._ind_active = ind_active
        self.default_key = default_key

    @property
    def model(self):
        """获取默认model（active形式）

        Notes
        -----
        默认活跃的数据为：
        1. models的第一个键
        2. 若models为空，则为kwargs的第一个键
        3. 若kwargs也为空，无选中的激活数据，则为active_cells
        """
        return self.get_active_model()

    @property
    def mesh(self):
        return self._mesh

    @property
    def active_cells(self):
        return self._ind_active

    @active_cells.setter
    def active_cells(self, val):
        self._ind_active = val

    @property
    def n_active_cells(self):
        if self._ind_active.dtype == bool:
            return np.count_nonzero(self._ind_active)
        else:
            return self._ind_active.shape[0]

    @property
    def n_cells(self):
        return self.mesh.n_cells

    @property
    def default_key(self):
        return self._default_key

    @default_key.setter
    def default_key(self, k):
        if k is not None:
            assert k in self._models, f'Cannot set `default_key` to `{k}` which does not exist.'
        self._default_key = k

    @property
    def has_default_model(self):
        return self.default_key is not None

    @property
    def has_only_default_model(self):
        return len(self) == 1 and self.has_default_model

    @property
    def cell_centers(self):
        return self.mesh.cell_centers

    @property
    def active_cell_centers(self):
        return self.cell_centers[self.active_cells]

    def set_active_model(self, key):
        self.default_key = key

    def clear_active_model(self):
        self.default_key = None

    def get_raw_model(self, key):
        """直接获取模型数组，不保证返回结果是有效模型或完整模型
        """
        return self._models[key]

    def keys(self):
        return self._models.keys()

    def get_active_model(self, key=None, **kwargs):
        if key is None:
            if self.has_default_model:
                key = self.default_key
            else:
                return np.ones(self.n_active_cells, dtype=bool)

        return self.map_to_active(self.get_raw_model(key), **kwargs)

    def is_active_model(self, model):
        return model.shape[0] == self.n_active_cells

    def is_complete_model(self, model):
        return model.shape[0] == self.n_cells

    def is_subset(self, mask):
        mask = self.check_complete_mask(mask)
        return np.all(self.active_cells | mask == self.active_cells)

    def get_complete_model(self, key=None, **kwargs):
        if key is None:
            if self.has_default_model:
                key = self.default_key
            else:
                return self.active_cells

        return self.map_to_complete(self.get_raw_model(key), **kwargs)

    def _unsupported_model_exception(self, model):
        return ValueError(f'Unsupported model size {model.shape},'
                          f' which is neither complete model ({self.n_cells},)'
                          f' nor active model ({self.n_active_cells},)')

    def map_to_complete(self, model: ArrayLike, dtype=None, fill_inactive=None, copy=True):
        """从有效网格模型或完整网格模型映射到完整网格模型

        Parameters
        ----------
        model
            待映射的模型
        dtype
            映射模型的值类型
        fill_inactive
            映射到完整网格时非活跃网格的填充值，若指定填充值不为None，则一定会发生拷贝
        copy
            若为`False`，则输入模型满足所有条件时不进行拷贝，否则保证返回值与输入不共享内存空间

        Returns
        -------
        ret
            在完整网格下的模型数组
        """
        from metalpy.scab.modelling import Scene

        if self.n_active_cells == self.n_cells:
            fill_inactive = None

        if fill_inactive is not None:
            copy = True
            ftype = type(fill_inactive)
            if dtype is None:
                dtype = ftype
            else:
                assert np.can_cast(ftype, dtype), f'Cannot cast fill value with type `{ftype}` to {dtype}.'

        model = np.asarray(model)
        if self.is_complete_model(model):
            ret = model
            if dtype is not None:
                ret = ret.astype(dtype, copy=copy)
                copy = False

            if copy:
                ret = ret.copy()

            if fill_inactive is not None:
                ret[~self.active_cells] = fill_inactive
        elif self.is_active_model(model):
            base_type = dtype or model.dtype
            if fill_inactive is None:
                inactive_val = Scene.get_inactive_value_for_type(base_type)
            else:
                inactive_val = fill_inactive
            ret = np.full(self.n_cells, inactive_val, dtype=base_type)
            ret[self.active_cells] = model
        else:
            raise self._unsupported_model_exception(model)

        return ret

    def map_to_active(self, model: ArrayLike, dtype=None, copy=True):
        model = np.asarray(model)
        if self.is_active_model(model):
            ret = model
        elif self.is_complete_model(model):
            ret = model[self.active_cells]
        else:
            raise self._unsupported_model_exception(model)

        if dtype is not None:
            ret = ret.astype(dtype, copy=copy)
            copy = False

        if copy:
            ret = ret.copy()

        return ret

    def check_complete_mask(self, mask: ArrayLike):
        mask = np.asarray(mask)
        assert mask.ndim == 1, 'Indices must 1D vector.'

        if mask.dtype == bool:
            assert mask.shape[0] == self.n_cells, \
                f'Invalid complete mask (expected {self.n_cells}, got {mask.shape[0]}).'
            ret = mask
        else:
            # assumed to be fancy indexing
            ret = np.full(self.n_cells, False, dtype=bool)
            ret[mask] = True

        return ret

    def check_active_mask(self, mask: ArrayLike):
        mask = np.asarray(mask)
        assert mask.ndim == 1, 'Indices must 1D vector.'

        if mask.dtype == bool:
            assert mask.shape[0] == self.n_active_cells, \
                f'Invalid active mask (expected {self.n_active_cells}, got {mask.shape[0]}).'
            ret = mask
        else:
            # assumed to be fancy indexing
            ret = np.full(self.n_active_cells, False, dtype=bool)
            ret[mask] = True

        return ret

    def is_active_mask(self, mask: ArrayLike):
        mask = np.asarray(mask)
        if mask.ndim != 1:
            return False
        if mask.dtype == bool:
            return self.n_active_cells == mask.shape[0]
        else:
            # assumed to be fancy indexing
            return mask.max() < self.n_active_cells

    def is_complete_mask(self, mask: ArrayLike):
        mask = np.asarray(mask)
        if mask.ndim != 1:
            return False
        if mask.dtype == bool:
            return self.n_cells == mask.shape[0]
        else:
            # assumed to be fancy indexing
            return mask.max() < self.n_cells

    def get(self, key, default=None):
        return self._models.get(key, default)

    def active_models(self):
        for key in self:
            yield key, self.get_active_model(key)

    def complete_models(self):
        for key in self:
            yield key, self.get_complete_model(key)

    def __iter__(self):
        yield from self._models

    def __len__(self):
        return len(self._models)

    def __getitem__(self, item):
        return self.get_active_model(item)

    def __setitem__(self, item, model: ArrayLike):
        model = np.asarray(model)
        assert model.shape[0] in (self.n_active_cells, self.n_cells), \
            f'`ModelledMesh` only accepts models matching active cells or all mesh cells.' \
            f' Got model {item} with size {model.shape[0]}, expected {self.n_active_cells} or {self.n_cells}.'
        self._models[item] = model

    def __contains__(self, item):
        return item in self._models

    def plot(self, *, scalars=True, prune: Literal[True, False] = True, **kwargs):
        if scalars is True:
            active_scalars = self.default_key
            if active_scalars is not None:
                scalars = active_scalars

        poly = self.to_polydata(scalars=scalars, prune=prune)

        poly.plot(**kwargs)

    @overload
    def to_polydata(self,
                    scalars: str | Iterable[str] | ArrayLike | bool = True,
                    extra_models: dict[str, ArrayLike] | None = None,
                    fill_inactive=None,
                    *,
                    prune: Literal[False] = False,
                    **kwargs: ArrayLike) -> 'pv.RectilinearGrid': ...

    @overload
    def to_polydata(self,
                    scalars: str | Iterable[str] | ArrayLike | bool = True,
                    extra_models: dict[str, ArrayLike] | None = None,
                    fill_inactive=None,
                    *,
                    prune: Literal[True],
                    **kwargs: ArrayLike) -> 'pv.UnstructuredGrid': ...

    def to_polydata(self,
                    scalars: str | Iterable[str] | ArrayLike | bool = True,
                    extra_models: dict[str, ArrayLike] | None = None,
                    fill_inactive=None,
                    prune: bool = True,
                    **kwargs: ArrayLike) -> 'pv.RectilinearGrid' | 'pv.UnstructuredGrid':
        """导出模型网格为PyVista模型

        Parameters
        ----------
        scalars
            需要导出的模型名称。
            True代表全部导出，False代表不导出包含的模型。
            通过字符串或字符串列表指定需要导出的模型。
            或者指定为数组以导出该指定模型。
        extra_models
            额外需要绑定的模型
        kwargs
            额外需要绑定的模型，但是以kwargs风格传入
        fill_inactive
            用于填充非活跃网格的值，为None时默认使用模型对应类型的默认值（一般为0，参见Scene.INACTIVE_XXX）
        prune
            从导出的 `RectilinearGrid` 中筛选活跃网格（会导致返回类型变为 `UnstructuredGrid` ）

        Returns
        -------
        ret
            包含指定的需要导出的模型和额外模型的pv.PolyData实例

        Notes
        -----
        虽然名字为to_polydata，但他的返回值并不是 `pv.PolyData` 而是 `pv.RectilinearGrid` 或 `pv.UnstructuredGrid`
        """
        import pyvista as pv
        from metalpy.scab.modelling import Scene

        default_key_active_cells = 'ACTIVE'
        active_scalars = None

        models = {}
        if not is_numeric_array(scalars):
            if scalars is True:
                active_scalars = self.default_key
            elif isinstance(scalars, str):
                active_scalars = scalars
            scalars = self._check_scalars(scalars)

            for scalar in scalars:
                if active_scalars is None:
                    active_scalars = scalar
                models[scalar] = self._models[scalar]
        else:
            # assumed to be model array
            key_model = 'model'

            active_scalars = ensure_set_key(
                models,
                key_model,
                scalars
            )

        if extra_models is not None:
            models.update(extra_models)
            if active_scalars is None:
                active_scalars = get_first_key(extra_models)

        models.update(kwargs)
        if active_scalars is None:
            active_scalars = get_first_key(kwargs)

        key_active_cells = ensure_set_key(
            models,
            default_key_active_cells,
            self._ind_active  # 保证为complete_model
        )
        if key_active_cells != default_key_active_cells:
            warnings.warn(f'Conflicted key with default key `{default_key_active_cells}` detected,'
                          f' renaming it to `{key_active_cells}`.')

        for key in models:
            model = models[key]
            model = self.map_to_complete(model, fill_inactive=fill_inactive)
            models[key] = model

        if active_scalars is None:
            active_scalars = key_active_cells

        ret: pv.RectilinearGrid = Scene.mesh_to_polydata(self.mesh, models)
        ret.set_active_scalars(active_scalars)

        if prune:
            return ret.extract_cells(self.active_cells)

        return ret

    def extract(self,
                indices: ArrayLike,
                scalars: str | Iterable[str] | bool = True,
                shallow=False) -> 'ModelledMesh':
        """通过网格mask提取子网格

        Parameters
        ----------
        indices
            有效网格坐标系下的mask（bool数组或下标数组）。
        scalars
            需要保留到新ModelledMesh实例的一个或多个模型的键名。
            若为True，则保留所有模型，若为False，则只提取网格不提取模型。
        shallow
            指示是否浅拷贝模型，否则在可能的时候会使用原模型的视图，某些情况可能导致意外的修改

        Returns
        -------
        ret
            包含新的有效网格掩码的模型网格

        Notes
        -----
        如果传入完整网格坐标系的mask，则只提取其和当前有效网格的交集，保证结果为当前有效网格的子集、
        """
        if self.is_active_mask(indices):
            indices = self.check_active_mask(indices)
            new_active_cells = self.map_to_complete(indices)
        else:
            new_active_cells = self.check_complete_mask(indices)
            new_active_cells &= self.active_cells
            indices = self.map_to_active(new_active_cells)

        scalars = self._check_scalars(scalars)
        models = {}

        for model_key in scalars:
            model = self.get_raw_model(model_key)
            if self.is_active_model(model):
                model = model[indices]
            else:
                if not shallow:
                    model = model.copy()
            models[model_key] = model

        ret = ModelledMesh(self.mesh, models=models, ind_active=new_active_cells)

        if self.default_key in ret:
            ret.set_active_model(self.default_key)

        return ret

    def reactivate(self,
                   indices,
                   scalars: str | Iterable[str] | bool = True,
                   shallow=False,
                   ) -> 'ModelledMesh':
        """给定新的有效网格掩码，重新定义模型网格

        Parameters
        ----------
        indices
            全局坐标系下的新有效网格掩码，用于重新生成网格
        scalars
            需要保留到新ModelledMesh实例的一个或多个模型的键名。
            若为True，则保留所有模型，若为False，则只提取网格不提取模型。
        shallow
            指示是否浅拷贝模型，否则在可能的时候会使用原模型的视图，某些情况可能导致意外的修改

        Returns
        -------
        ret
            包含新的有效网格掩码的模型网格

        Notes
        -----
            由于目前无效网格值会被定义为0，所以新的mask和旧有mask交集之外的值会是0。
            如果shallow为真，则不保证新旧mask交集之外的值，因此在多级reactivate调用下，shallow模式可能会导致大量不可预测的结果
        """
        new_active_cells = self.check_complete_mask(indices)

        is_subset = self.is_subset(new_active_cells)
        if is_subset:
            return self.extract(self.map_to_active(new_active_cells), scalars=scalars, shallow=shallow)
        else:
            ret = ModelledMesh(self.mesh, ind_active=new_active_cells)
            scalars = self._check_scalars(scalars)

            for model_key in scalars:
                model = self.get_raw_model(model_key)

                if not self.is_complete_model(model):
                    from metalpy.scab.modelling import Scene

                    warnings.warn(f'Reactivated mesh with model for active cells.'
                                  f' May lead to unexpected result in non-overlapping cells,'
                                  f' where inactive cells are filled with default value'
                                  f' `{Scene.get_inactive_value_for_type(model.dtype)}`')

                    model = self.map_to_complete(model)
                else:
                    if not shallow:
                        model = model.copy()

                ret[model_key] = model

            if self.default_key in ret:
                ret.set_active_model(self.default_key)

            return ret

    def rebind(
        self,
        model: str | ArrayLike,
        scalars: str | Iterable[str] | bool = False,
        shallow=False
    ) -> 'ModelledMesh':
        """基于将网格重新绑定到model模型上或model作为字符串指定的模型上

        Parameters
        ----------
        model
            需要重新绑定的新模型或model作为字符串指定的需要重新绑定的模型
        scalars
            需要一同提取到新ModelledMesh实例的模型名。
            若scalars为`False`，则新实例仅包含model指定的模型。
            若scalars为`True`，则新实例包含model指定的模型和当前实例的全部模型。
            若scalars为字符串或字符串集合，则新实例包含model指定的模型和scalars指定的模型。
        shallow
            指示是否浅拷贝模型，否则在可能的时候会使用原模型的视图，某些情况可能导致意外的修改

        Returns
        -------
        ret
            包含新的有效网格掩码的模型网格
        """
        from metalpy.scab.modelling import Scene

        scalars = set(self._check_scalars(scalars))

        if isinstance(model, str):
            if model not in scalars:
                scalars.add(model)
            ref = self.get_raw_model(model)
            mask = Scene.is_active(ref)
        else:
            mask = Scene.is_active(model)

        if self.is_active_model(mask):
            ret = self.extract(mask, scalars=scalars, shallow=shallow)
        else:
            ret = self.reactivate(mask, scalars=scalars, shallow=shallow)

        if not isinstance(model, str):
            if self.is_active_model(mask):
                # mask是self下的有效网格模型，意味着其不是ret下的有效网格模型
                # 因此model不能直接继承给ret，而需要重新映射为ret下的有效网格模型
                # 例如：假设self总网格3000个，有效网格1000个，ret含有效网格500个，
                # model作为self的有效网格模型，长度为1000，
                # 而ret只接受500或3000两种模型格式，因此需要重新映射到长度为500的格式
                model = model[mask]
            model = ensure_set_key(ret, Object.DEFAULT_KEY, model)

        ret.set_active_model(model)

        return ret

    def _check_scalars(self, scalars) -> Iterable[str]:
        if scalars is True:
            return self.keys()
        elif scalars is False:
            return tuple()
        elif isinstance(scalars, str):
            return (scalars,)
        else:
            return scalars



from .function_context import FunctionTermination, ParameterModification
from .replaces import Replaces
from .utils import wrap_method_with_target, update_params


class Before(Replaces):
    def __init__(self, target, nest=None):
        super().__init__(target, keep_orig='__orig_fn__', nest=nest, force_unbound=True)

    def __call__(self, func):
        func, is_method = wrap_method_with_target(self.nest, func)

        def wrapper(*args, __orig_fn__=None, **kwargs):
            ret = func(*args, **kwargs)
            if isinstance(ret, ParameterModification):
                args, kwargs = update_params(__orig_fn__, args, kwargs, ret.new_args)
            elif isinstance(ret, FunctionTermination):
                return ret.ret
            return __orig_fn__(*args, **kwargs)

        return super().__call__(wrapper)


def before(target, nest=None):
    """将代码注入到目标函数前执行

    Parameters
    ----------
    target
        待注入的目标函数
    nest
        手动指定替换对象的所在空间

    Notes
    -----
        return terminate() 或 return terminate_with(*args) 可以指示提前终止函数

    See Also
    --------
        after: 将代码注入到函数之后执行
        replaces: 替换目标函数
        terminate or terminate_with: 指示以无返回值或指定返回值提前终止函数
    """
    return Before(target, nest=nest)



from typing import Iterable, Union

import numpy as np

from metalpy.utils.dhash import dhash
from metalpy.utils.bounds import Bounds
from . import Shape3D


class Cuboid(Shape3D):
    def __init__(self,
                 origin: Iterable = None,
                 end: Iterable = None,
                 corner: Iterable = None,
                 corner2: Iterable = None,
                 size: Union[Iterable, float] = None,
                 center: Iterable = None,
                 bounds: Iterable = None,
                 no_corner_adjust=False):
        """通过立方体的中心、长度、角点中任意两个参数或单独使用边界来定义立方体

        Parameters
        ----------
        origin, end
            立方体的角点坐标
        corner, corner2
            立方体的角点坐标，用于兼容，建议使用origin和end
        size
            立方体的长宽高
        center
            中心点坐标
        bounds
            立方体的边界，当定义时，无视其他参数，[x0, x1, y0, y1, z0, z1]
        no_corner_adjust
            是否拒绝调整角点坐标，默认为假，出现负数边长时，会自动调整角点坐标，若为真，则抛出异常
        """
        super().__init__()

        if origin is None:
            origin = corner

        if end is None:
            end = corner2

        if bounds is None and np.r_[
            origin is not None,
            end is not None,
            center is not None,
            size is not None,
        ].sum() != 2:
            raise ValueError("Bounds or exactly two of"
                             " `center`, `lengths`, `origin`, `end`"
                             " must be specified.")

        if origin is not None:
            origin = np.array(origin)

        if size is not None:
            size = np.array(size)
            if size.size == 1:
                size = np.ones(3) * size

        if bounds is not None:
            origin = np.array(bounds[::2])
            size = np.array(bounds[1::2]) - origin
        elif origin is not None:  # 和end或center，计算size
            if end is not None:
                size = np.asarray(end) - np.asarray(origin)
            elif center is not None:
                size = 2 * (np.asarray(center) - np.asarray(origin))
        elif end is not None:  # 和center或size，计算origin
            if center is not None:
                size = 2 * (np.asarray(end) - np.asarray(center))
                origin = np.asarray(end) - size
            elif size is not None:
                origin = np.asarray(end) - np.asarray(size)
        elif size is not None:  # 和center，计算origin
            if center is not None:
                origin = np.asarray(center) - size / 2

        if np.any(size < 0):
            if no_corner_adjust:
                raise ValueError("Negative length detected.")
            # 纠正负数边长
            origin[size < 0] += size[size < 0]
            size = np.abs(size)

        self.origin = origin
        self.lengths = size

    def to_prism(self):
        from .prism import Prism
        return Prism(pts=self.pts, z0=self.z0, z1=self.z1)

    def do_place(self, mesh_cell_centers, progress):
        return np.full(len(mesh_cell_centers), True)

    @property
    def x0(self): return self.origin[0]

    @property
    def x1(self): return self.origin[0] + self.lengths[0]

    @property
    def y0(self): return self.origin[1]

    @property
    def y1(self): return self.origin[1] + self.lengths[1]

    @property
    def z0(self): return self.origin[2]

    @property
    def z1(self): return self.origin[2] + self.lengths[2]

    @x0.setter
    def x0(self, val):
        self.lengths[0] = self.x1 - val
        self.origin[0] = val

    @x1.setter
    def x1(self, val): self.lengths[0] = val - self.x0

    @y0.setter
    def y0(self, val):
        self.lengths[1] = self.y1 - val
        self.origin[1] = val

    @y1.setter
    def y1(self, val): self.lengths[1] = val - self.y0

    @z0.setter
    def z0(self, val):
        self.lengths[2] = self.z1 - val
        self.origin[2] = val

    @z1.setter
    def z1(self, val): self.lengths[2] = val - self.z0

    @property
    def corner(self): return self.origin

    @property
    def end(self): return self.origin + self.lengths

    @property
    def pts(self):
        """兼容Prism接口"""
        return np.asarray([
            [self.x0, self.y0],
            [self.x0, self.y1],
            [self.x1, self.y1],
            [self.x1, self.y0],
        ])

    @property
    def direction(self):
        """
        获取横梁最长轴方向

        Returns
        -------
        axis
            0, 1, 2代表x, y, z
        """
        return np.argmax(self.lengths)

    def __dhash__(self):
        return dhash(super().__dhash__(), *self.origin, *self.lengths)

    def do_clone(self, deep=True):
        return Cuboid(
            origin=self.origin.copy(),
            size=self.lengths.copy()
        )

    @property
    def local_bounds(self):
        return Bounds(np.c_[self.origin, self.origin + self.lengths].ravel())

    def to_local_polydata(self):
        import pyvista as pv

        return pv.Cube(bounds=np.asarray(self.local_bounds))

    @property
    def volume(self):
        return np.prod(self.lengths)

    @property
    def area(self):
        return 2 * (self.lengths * np.roll(self.lengths, 1)).sum()



from __future__ import annotations

from typing import Union

from metalpy.mexin import Patch
from metalpy.scab import Tied
from metalpy.scab.distributed.policies import Distributable
from metalpy.utils.type import get_params_dict


class Demaged(Patch, Distributable):
    Priority = Tied.Priority + 1  # 保证在Tied后面注入mixin

    def __init__(
            self,
            factor=None,
            method=None,
            compressed_size: Union[int, float, None] = None,
            deterministic: Union[bool, str] = True,
            quantized: bool = False,
            symmetric: bool | None = None,
            progress: bool = False,
            **kwargs
    ):
        """为 `Simulation3DIntegral` 添加退磁效应计算

        Parameters
        ----------
        factor
            三轴退磁系数，如果为None，则采用数值方法求解
        method
            采用的数值求解算法，参考 `Demagnetization.methods`。默认为None，自动选择。
            可选选项：
                - Demagnetization.Compressed
                - Demagnetization.Seperated
                - Demagnetization.Integrated
        compressed_size
            压缩表尺寸，只针对`Demagnetization.Compressed`有效，用于指定压缩后的核矩阵哈希表最大大小，应尽可能小以满足内存要求。
            浮点数代表按完整核矩阵大小的比例设置，整数代表直接设置尺寸。
            如果大小不足或过大内部会给出提示信息
        deterministic
            采用确定性模式并行哈希，默认为True，牺牲一定计算效率和空间确保核矩阵正确。
            指定为CompressedForward.Optimal时直接采用unique函数计算理想压缩结果
        quantized
            采用量化模式压缩核矩阵，默认为False。
            True则启用压缩，使用更小位宽来存储核矩阵，减少一定的空间需求，略微牺牲计算效率。
            （可能会报要求尺寸为2的幂的性能警告）
        symmetric
            指示是否启用对称模式（当网格为规则网格时可以启用）。
            True则在检测到网格符合条件时启用对称模式，以对称矩阵形式存储核矩阵，减少一定的空间需求。
            False则禁用对称模式，无论是否符合条件。
            在CPU模式下会对性能造成较大影响，因此默认不启用。
            在GPU模式下性能影响较小，因此默认启用。
        progress
            是否输出求解进度条，默认为False不输出
        """
        super().__init__()
        self.params = get_params_dict(
            factor=factor,
            method=method,
            compressed_size=compressed_size,
            deterministic=deterministic,
            quantized=quantized,
            symmetric=symmetric,
            progress=progress,
            **kwargs
        )

    def apply(self):
        from SimPEG.potential_fields.magnetics import Simulation3DIntegral
        from metalpy.scab.demag.demaged_mixin import DemagedMixin

        self.add_mixin(Simulation3DIntegral, DemagedMixin, **self.params)

    def bind_context(self, ctx):
        super().bind_context(ctx)



from .array_ops import ArrayType, OperatorDispatcher

from .is_array_like import is_array_like
from .concat import concat
from .get_column import get_column
from .get_columns_except import get_columns_except
from .get_row import get_row
from .ustack import ustack, cstack, rstack



from .worker_context import WorkerContext


class QueueLikeWorkerContext(WorkerContext):
    def __init__(self, queue=None):
        self.queue = queue

    def fire(self, event, *args, **kwargs):
        self.queue.put((event, (args, kwargs)))



from __future__ import annotations

import contextlib
import os
import warnings
from io import BytesIO
from multiprocessing.pool import ThreadPool
from pathlib import Path
from threading import Lock
from typing import Iterable, Generator

import numpy as np
import requests
import tqdm
from PIL import Image

from metalpy.utils.file import make_cache_directory_path, ensure_filepath, PathLike
from metalpy.utils.bounds import Bounds
from metalpy.utils.dhash import dhash
from metalpy.utils.path import pathencode

from .constants import WebMercator
from .geo_image import GeoImage
from .geo_image_ref_system import GeoImageRefSystem
from .sources.tile_map_source import TileMapSource
from .sources.tile_locator import TileLocator


class MapDownloader:
    """
    References
    ----------
    https://github.com/cutecore/google_map_satellite_download
    https://github.com/Icemap/go-map-downloader
    """
    def __init__(self, map_source: TileMapSource, n_threads=50, max_retries=3):
        # WMTS only for now
        self.map_source = map_source
        self.pool = ThreadPool(n_threads)

        self.max_retries = max_retries

        self.tile_locators: list[TileLocator] | None = None
        self.running = False
        self.session: requests.Session | None = None
        self.canvas_lock = Lock()
        self.progress = None

    def download(self, west_lon=None, east_lon=None, south_lat=None, north_lat=None,
                 bounds=None,
                 mercator_bounds=None,
                 levels=(18,),
                 combine: bool | PathLike | None = None,
                 crop: bool | PathLike | None = True,
                 geotiff: PathLike | None = None,
                 cache: bool | PathLike | None = None) -> GeoImage:
        """下载区域内地图瓦片，并进行组合、裁剪得到参数指定区域的地图数据

        Parameters
        ----------
        west_lon, east_lon, south_lat, north_lat
            下载地图的西、东、南、北边界WGS 84坐标
        bounds
            下载地图的边界WGS 84坐标，以[lon_min, lon_max, lat_min, lat_max]顺序提供
        mercator_bounds
            下载地图的边界Pseudo-Mercator坐标，以[xmin, xmax, ymin, ymax]顺序提供
        levels
            需要下载的地图等级
        combine
            下载的全部关联瓦片拼接后的存储路径
        crop
            下载的全部关联瓦片拼接后按边界裁剪结果的存储路径
        geotiff
            瓦片结果（是否裁剪后取决于是否crop）绑定地理坐标后的存储路径
        cache
            瓦片及中间结果的缓存路径

        Returns
        -------
        map_data
            处理后的地图数据

        Notes
        -----
        cache缓存路径具体为'{cache_path}/{map_source_name}'
        """
        ret = None

        if not combine and not crop:
            assert geotiff, 'At least one of `combine`, `crop` or `geotiff` must be specified.'
            combine = True

        if bounds is not None:
            bounds = Bounds(bounds)
        elif mercator_bounds is not None:
            # 因为地图源的坐标转换以经纬度为单位，因此仍然需要一份WGS 84边界坐标
            mercator_bounds = Bounds(mercator_bounds)
            bounds = WebMercator.pseudo_mercator_to_wgs84(mercator_bounds[0:2], mercator_bounds[2:4])
            bounds = np.concatenate(bounds).view(Bounds)
        else:
            bounds = Bounds([west_lon, east_lon, south_lat, north_lat])

        if mercator_bounds is None:
            tmb_xy = WebMercator.wgs84_to_pseudo_mercator(bounds[0:2], bounds[2:4])
            true_mercator_bounds = np.concatenate(tmb_xy)
        else:
            true_mercator_bounds = np.asarray(mercator_bounds)
        true_mercator_bounds = Bounds(true_mercator_bounds)

        bounds.origin = self.map_source.warp_coordinates(bounds.origin)
        bounds.end = self.map_source.warp_coordinates(bounds.end)

        mb_xy = WebMercator.wgs84_to_pseudo_mercator(bounds[0:2], bounds[2:4])
        mercator_bounds = np.concatenate(mb_xy).view(Bounds)

        warped = not np.allclose(mercator_bounds, true_mercator_bounds)

        if not isinstance(levels, Iterable):
            levels = (levels,)

        with self.activate_pool():
            for level in levels:
                tiles_mercator_bounds = WebMercator.to_bounding_tiles_mercator_bounds(mercator_bounds, level)
                ref_system: GeoImageRefSystem | None = None
                tiles = self.download_geo_tiles(
                    mercator_bounds, level,
                    cache=cache
                )
                for tile in tiles:
                    if ref_system is None:
                        ref_system = tile.ref_system
                        if combine:
                            canvas = ref_system.create_image(
                                geo_bounds=tiles_mercator_bounds
                            )
                        else:  # 上文保证crop和combine至少有一个为True，因此此处crop=True
                            canvas = ref_system.create_image(
                                geo_bounds=mercator_bounds
                            )
                    canvas.paste(tile)

                ret = canvas

                if combine:
                    combined = canvas
                    if not isinstance(combine, bool):
                        saved_path = MapDownloader.encode_filename_with_level(
                            combine, level, levels, 'combined'
                        )
                        combined.save(saved_path)

                if crop:
                    if combine:
                        cropped = combined.crop(geo_bounds=mercator_bounds)
                    else:
                        cropped = canvas

                    ret = cropped

                    if not isinstance(crop, bool):
                        saved_path = MapDownloader.encode_filename_with_level(
                            crop, level, levels, 'cropped'
                        )
                        cropped.save(saved_path)

                # combine和crop输出png图像，与地理坐标系无关，因此坐标纠正可以在之后进行
                if warped:
                    # true_mercator_bounds为标准WGS 84坐标系下的地图边界
                    # mercator_bounds为地图源转换后坐标系下的地图边界
                    # 此处假设坐标转换前后区域尺度整体为线性变换，因此直接平移缩放转换到WGS 84下的地图边界
                    checks = []

                    # 尝试转换原点
                    ref_system.origin -= mercator_bounds.origin - true_mercator_bounds.origin

                    # 检查边界对齐情况
                    if crop:
                        edge_true_mercator_bounds = ref_system.to_edge_geo_bounds(true_mercator_bounds)
                        delta_bounds = np.abs(canvas.edge_geo_bounds - edge_true_mercator_bounds)
                        pixel_biases = delta_bounds.as_corners() / canvas.unit_size
                        if np.any(pixel_biases > 1):
                            checks.append(f'misaligned boundary ({pixel_biases.max():.2}px)')

                    # 计算源和目标区域缩放因子
                    scale_factor = true_mercator_bounds.extent / mercator_bounds.extent

                    # 检查区域尺度差异
                    relative_extent_diff = np.abs(scale_factor - 1)
                    if np.any(relative_extent_diff > 0.01):
                        checks.append(f'large scale difference ({relative_extent_diff.max():.2%})')

                    # 尝试转换坐标轴尺度
                    ref_system.unit_size *= scale_factor

                    if len(checks) > 0:
                        warnings.warn(f'{" and ".join(checks).capitalize()} detected.'
                                      f' due to coordinates transformation by `{self.map_source}`.'
                                      f' Consider using other map sources instead.')

                if geotiff:
                    geotiff_save_path = MapDownloader.encode_filename_with_level(
                        geotiff, level, levels, 'geotiff'
                    )

                    ret.save_geo_tiff(geotiff_save_path)

        return ret

    def download_geo_tiles(self,
                           mercator_bounds,
                           level,
                           cache: bool | PathLike | None = None,
                           true_tiles_mercator_bounds=None,
                           ref_system=None) -> Generator[GeoImage]:
        """下载Tile地图并转换为GeoTIFF形式

        Parameters
        ----------
        mercator_bounds
            Pseudo-Mercator边界
        level
            缩放等级
        cache
            缓存/缓存路径
        true_tiles_mercator_bounds
            边界坐标变换前的Pseudo-Mercator边界
        ref_system
            参考系（原点需为裁剪前瓦片地图区域左下角）

        Returns
        -------
        iterable[geo_tile]
            带坐标位置信息的Tile序列

        See Also
        --------
        MapDownloader.download : 地图下载主函数

        Notes
        -----
        某些地图源由于坐标系不一致，包含边界坐标变换步骤，需要调用者手动执行（参考MapDownloader.download方法）
        """
        tiles = list(WebMercator.iter_tiles(mercator_bounds, level))
        force_cache, cache_path = self.get_cache_path(cache)

        _cache_path = None
        if force_cache or len(tiles) > 100:
            _cache_path = cache_path

        tile_bounds = WebMercator.pseudo_mercator_bounds_to_tile(mercator_bounds, level)
        tiles_mercator_bounds = WebMercator.tile_bounds_to_pseudo_mercator(tile_bounds, level)

        if true_tiles_mercator_bounds is None:
            true_tiles_mercator_bounds = tiles_mercator_bounds
        else:
            true_tiles_mercator_bounds = Bounds(true_tiles_mercator_bounds)

        tile_extent = true_tiles_mercator_bounds.extent
        tile_count = tile_bounds.extent + 1
        col0, row1 = tile_bounds.xmin, tile_bounds.ymax

        tile_size = None

        for col, row, tile in self.download_tiles(tiles, level, _cache_path):
            if tile_size is None:
                tile_size = tile.size

            if ref_system is None:
                combined_image_size = (
                    tile_size[0] * tile_count[0],
                    tile_size[1] * tile_count[1]
                )
                ref_system = GeoImageRefSystem.of_unit_edge_bounds(
                    true_tiles_mercator_bounds.origin,
                    tile_extent / combined_image_size,
                    crs=WebMercator.PseudoMercator,
                    flip_y=True
                )

            geotile = ref_system.map_image(tile, offset=(
                (col - col0) * tile_size[0],
                (row1 - row) * tile_size[1]
            ))

            yield geotile

    def get_cache_path(self, cache):
        if cache is not None and not isinstance(cache, bool):
            cache_path = Path(cache)
            force_cache = True
        else:
            dirname = pathencode(repr(self.map_source))
            try:
                cache_path = make_cache_directory_path(
                    'map_downloader',
                    dirname
                )
            except FileNotFoundError:  # 存在可能性文件名过长？
                cache_path = make_cache_directory_path(
                    'map_downloader',
                    dhash(dirname).hexdigest()
                )
            force_cache = cache

        return force_cache, cache_path

    def download_tiles(self, tiles, level, cache_path):
        futures = []
        tile_source_index = 0
        progress = tqdm.tqdm(total=len(tiles))

        for x, y in tiles:
            futures.append(self.pool.apply_async(self.download_tile, (
                tile_source_index, x, y, level, cache_path
            )))
            tile_source_index = self.next_tile_source_index(tile_source_index)

        for (col, row), future in zip(tiles, futures):
            tile = future.get()
            progress.update(1)
            yield col, row, tile

        progress.close()

    @staticmethod
    def encode_filename_with_level(path, level, levels, source):
        raw_path = os.fspath(path)
        if '{level}' in raw_path:
            path = Path(raw_path.format(level=level))
            if path.is_dir():
                path = path / f'{source}.png'
        else:
            path = Path(path)

            if len(levels) == 1:
                if path.is_dir():
                    path = path / f'{source}.png'
            else:
                if path.is_dir():
                    path = path / f'{source}.{level}.png'
                else:
                    example = path.with_suffix(f'.{{level}}{path.suffix}')
                    path = path.with_suffix(f'.{level}{path.suffix}')
                    warnings.warn(f'Multiple levels share same file name, renamed to `{path}`.'
                                  f' Try using a placeholder in path like `{example}` or path to a directory.')

        return path

    def next_tile_source_index(self, i):
        return (i + 1) % len(self.tile_locators)

    def download_tile(self, priority_source_index: int, col, row, level,
                      cache_path: Path | None = None):
        tile = None
        cache_file = None
        if cache_path is not None:
            cache_file = cache_path / f'{level}/{col}/{row}.png'

            if cache_file.exists() and os.path.getsize(cache_file) > 400:
                tile = Image.open(cache_file)
            else:
                ensure_filepath(cache_file)

        if tile is None:
            error_count = 0
            map_data = None
            while error_count < self.max_retries:
                priority_source_index = self.next_tile_source_index(priority_source_index)
                tile_source = self.tile_locators[priority_source_index]

                url = tile_source.query(col, row, level)

                try:
                    resp = self.session.get(url)
                    if resp.status_code == 200:
                        map_data = resp.content
                        break
                except requests.RequestException:
                    continue

                error_count += 1

            if map_data is None:
                raise RuntimeError(f'Failed to download map tile (Max retries {self.max_retries} exceeded).')

            tile = Image.open(BytesIO(map_data))

        if cache_path is not None:
            tile.save(cache_file)

        return tile

    @contextlib.contextmanager
    def activate_pool(self):
        self.running = True
        self.tile_locators = [t for t in self.map_source]
        self.session = requests.Session()

        try:
            yield
        finally:
            pass

        self.running = False
        self.tile_locators = None
        self.session = None
        self.progress = None



import inspect
import warnings

from .injectors import extends, replaces, after, before
from .mixin import Mixin, TaggedMethod
from .patch import Patch
from .sentinal import NO_MIXIN


class MixinManager:
    """会被注册为目标类的mixins属性，由其他mixin调用进行mixin的注入或访问
    """
    def __init__(self, target):
        self._mixins = {}
        self.target = target

    def add(self, mixin_type, *args, **kwargs):
        """将mixin绑定到关联对象上

        Parameters
        ----------
        mixin_type
            mixin类
        args
            mixin类的位置参数
        kwargs
            mixin类的关键字参数

        Notes
        -----
            使用类加参数的形式来定义mixin的主要原因是可能存在mixin中需要持有一些对象的引用导致无法序列化，
            比如mixin对象内可能持有一个tqdm进度条实例，但如果再包一层参数代理类又会过于冗杂，因此采用这种方式定义mixin
        """
        target = self.target
        obj = mixin_type(target, *args, **kwargs)

        # 获取mixin的所有方法
        methods = inspect.getmembers(obj, predicate=MixinManager._is_method)
        for name, method in methods:
            if name.startswith('__') or \
               f'_{mixin_type.__name__}__' in name or \
               name in Mixin.__dict__:
                # 跳过1. 魔术方法 2. 私有方法 3. post_apply等mixin类的方法
                continue

            obj.__dict__[name] = self.bind_method(method, name=name)

        self._mixins[mixin_type] = obj
        obj.post_apply(target)

    def get(self, mixin_type, sentinel=False):
        if mixin_type in self._mixins:
            return self._mixins[mixin_type]
        else:
            if sentinel:
                return NO_MIXIN
            else:
                return None

    def get_mixin_or_sentinel(self, mixin_type):
        return self.get(mixin_type, sentinel=True)

    def bind_method(self, method, name=None):
        # TODO: 引入注解来标记是否需要替换，或者标记替换别的目标
        # TODO: 标记是否需要保留原函数
        tag = TaggedMethod.Replaces
        keep_orig = False
        if name is None:
            name = method.__name__

        mixing_config = TaggedMethod.check(method)
        if mixing_config:
            tag = TaggedMethod.verify(mixing_config.tag)
            keep_orig = mixing_config.keep_orig
            if mixing_config.target is not None:
                # 用户指定的替换目标，覆盖之前获取到的名字
                name = mixing_config.target

        target = self.target
        target_method = getattr(target, name, None)

        if not target_method:
            # Replaces模式下如果不存在目标函数则隐式变为Extend，扩展新方法
            if tag != TaggedMethod.Replaces:
                warnings.warn(f'Trying to perform `{tag}`'
                              f' on non-existing method `{name}`.'
                              f' Consider using `{TaggedMethod.Replaces}` instead.')

            extension = method
            if keep_orig:
                # Extend时本不存在原函数，但用户想keep_orig，所以给一个空函数来make user happy
                warnings.warn(f'Trying to keep non-existing original method `{name}`.'
                              f' Populating with an empty function.')

                def extension(*args, **kwargs): ...

            method = extends(target, name)(extension)

            if keep_orig:
                # 用于触发真正的Replaces
                target_method = method

        if target_method:
            # 如果在目标存在同名方法则替换、前后插入
            if tag == TaggedMethod.Replaces:
                method = replaces(target_method, nest=target, keep_orig=keep_orig)(method)
            elif tag == TaggedMethod.Before:
                method = before(target_method, nest=target)(method)
            elif tag == TaggedMethod.After:
                method = after(target_method, nest=target)(method)

        return method

    @staticmethod
    def _is_method(func):
        return (
            inspect.ismethod(func)
            or inspect.isfunction(func)
            or TaggedMethod.check(func)
        )


class Mixed(Patch):
    def __init__(self, target):
        super().__init__()
        self.target = target
        self.pre_applied_mixins = []

    def mix(self, mixin_type, *args, **kwargs):
        self.pre_applied_mixins.append((mixin_type, args, kwargs))
        return self

    def apply(self):
        self.add_injector(after(self.target.__init__), self.apply_mixins)

    def apply_mixins(self, this, *_, **__):
        manager = getattr(this, 'mixins', None)
        if manager is None:
            manager = MixinManager(this)
            this.mixins = manager
        for m, args, kwargs in self.pre_applied_mixins:
            manager.add(m, *args, **kwargs)

    @property
    def priority(self):
        return -1



import functools

from metalpy.utils.object_path import get_nest as _get_nest, reassign_object_name, reassign_object_module


def create_replacement(func, orig, executor, qualname=None, module_name=None):
    if isinstance(func, property):
        # property不可以被构造新成员
        # 因此继承一个子类来实现引入原成员
        ret = PropertyReplacement(func, orig, executor)
    elif isinstance(executor.nest, type):
        # 类的成员函数必须是函数类型function才能在实例化时被识别并绑定self参数
        # 因此采用这个workaround来引入原函数
        func.repl_orig = orig
        func.repl_executor = executor
        ret = func
    else:
        ret = Replacement(func, orig, executor)

    if orig is None:
        if qualname is not None:
            reassign_object_name(ret, new_qualname=qualname)
        if module_name is not None:
            reassign_object_module(ret, module_name)
    else:
        functools.wraps(orig, updated=())(ret)

    return ret


class TemporaryReversion:
    def __init__(self, *repls):
        self.repls = repls
        self.replacements = []

    def __enter__(self):
        for repl in self.repls:
            self.replacements.append(repl.func)
            repl.repl_executor.rollback()

    def __exit__(self, exc_type, exc_val, exc_tb):
        for repl, replacement in zip(self.repls, self.replacements):
            repl.repl_executor(replacement)


def reverted(*repls):
    """临时还原某些替换
    例如类名劫持，有些类在调用基类构造函数时会采用类似于super(<class-name>, self).__init__(...)的形式，
    导致在被劫持状态下无法正常构造，这时需要进行临时的revert操作
    """
    return TemporaryReversion(*repls)


def get_orig(repl):
    if hasattr(repl, 'repl_orig'):
        return repl.repl_orig
    else:
        return None


def get_ancestor(repl):
    prev = repl
    orig = get_orig(repl)
    while orig is not None:
        prev = orig
        orig = get_orig(orig)
    return prev


def get_nest(repl):
    """适配Replacement的get_nest"""
    if hasattr(repl, 'repl_executor'):
        return repl.repl_executor.nest
    return _get_nest(repl)


def is_or_is_replacement(obj, other):
    return get_ancestor(obj) == get_ancestor(other)


class Replacement:
    def __init__(self, func, orig, executor):
        self.func = func
        self.repl_orig = orig
        self.repl_executor = executor

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)


class PropertyReplacement(property):
    def __init__(self, prop: property, orig, executor):
        super().__init__(prop.fget, prop.fset, prop.fdel, prop.__doc__)
        self.repl_orig = orig
        self.repl_executor = executor




import contextlib


@contextlib.contextmanager
def check_axis(ax, *, show=True, fig=None, figure=False, **fig_kw):
    """检查外部传入的axis是否有效，无效则基于 `fig_kw` 的参数调用 `plt.subplots` 创建fig和ax进行绘图

    Parameters
    ----------
    ax
        外部传入的matplotlib axis
    show
        指示如果是内部创建的figure，是否自动展示图像
    figure
        指示是否返回创建的figure，或者使用外部ax时返回None
    fig_kw
        用于创建新figure的参数，例如 `figsize`

    Notes
    -----
    如果 `figure=True`，则返回创建的figure和axis。

    如果 `figure=False`，则只返回axis。

    Examples
    --------
    >>> def plot_something(ax=None, show=True):
    >>>     with check_axis(ax, show=show, figsize=(5, 5)) as ax:
    >>>         ax.scatter([1, 2, 3, 4], [1, 6, 9, 16])

    >>> def plot_and_save(ax=None, fig=None):
    >>>     with check_axis(ax, show=False, figsize=(5, 5)) as fig, ax:
    >>>         ax.scatter([1, 2, 3, 4], [1, 6, 9, 16])
    >>>         if fig:
    >>>             fig.savefig('./screenshot.png')
    """
    from matplotlib import pyplot as plt

    if ax is None:
        fig, ax = plt.subplots(**fig_kw)

    if figure:
        yield fig, ax
    else:
        yield ax

    if show and fig is not None:
        fig.show()



import numpy as np


# return the hash of the given string in string format
def hash_str(*objs):
    flattened = []
    for obj in objs:
        if isinstance(obj, list):
            flattened.extend(obj)
        else:
            flattened.append(obj)

    flattened = tuple(flattened)
    hash_value = hash(flattened)

    return create_hash_str(hash_value)


def create_hash_str(hash_value, max_width=6):
    char_table = '0123456789abcdef'
    tablesize = len(char_table)
    hash_string = ''

    while hash_value != 0 and len(hash_string) < max_width:
        c = hash_value % tablesize
        hash_string = char_table[c] + hash_string
        hash_value = hash_value // tablesize

    return hash_string


def hash_numpy_array(arr: np.ndarray, n_samples=10, sparse=False):
    if not sparse:
        arr = arr.ravel()
        rand = np.random.RandomState(int(arr[len(arr) // 2]))

        # 加入shape作为参数防止类似np.ones(100)和np.ones(1000)的冲突
        return hash((arr.shape, *rand.choice(arr, min(len(arr), n_samples), replace=False),))
    else:
        # 稀疏的数组可能会在采样时碰撞概率较大
        # 但是涉及压缩操作可能会导致效率一定程度上降低
        import blosc2
        packed = blosc2.pack_array2(arr)

        return hash(packed)



from __future__ import annotations

import warnings

from metalpy.carto.coords import wgs2gcj
from .simple_tile_map_source import SimpleTileMapSource


class AMapSource(SimpleTileMapSource):
    LangDefault = None
    LangChs = 'zh-CN'
    LangEn = 'en'
    LayerBackground = 1
    LayerRoads = 2
    LayerLabels = 4
    LayerAll = LayerBackground | LayerRoads | LayerLabels

    SiteRd = r'https://webrd0[1234].is.autonavi.com/appmaptile\?x=\{x\}&y=\{y\}&z=\{z\}'  # deprecated
    SiteSt = r'https://webst0[1234].is.autonavi.com/appmaptile\?x=\{x\}&y=\{y\}&z=\{z\}'  # deprecated
    SiteWprd = r'https://wprd0[1234].is.autonavi.com/appmaptile\?x=\{x\}&y=\{y\}&z=\{z\}'

    def __init__(self,
                 satellite=False,
                 layers: int | None = None,
                 transparent_background=False,
                 site=None,
                 lang=None,
                 scl=None,
                 style=None
                 ):
        """高德地图数据源

        Parameters
        ----------
        """
        if site is None:
            site = AMapSource.SiteWprd

        if lang is None:
            lang = AMapSource.LangChs

        if layers is None and not satellite:
            layers = AMapSource.LayerAll

        components = [site]
        if lang is not None:
            components.append(f'lang={lang}')

        if style is None:
            if satellite:
                if layers is not None and layers != 0:
                    warnings.warn('AMap source does not support'
                                  ' download satellite map with other layers.')
                style = 6
            elif transparent_background:
                style = 8
            else:
                style = 7

        components.append(f'style={style}')
        if layers is not None:
            components.append(f'ltype={layers}')

        if scl is not None:
            components.append(f'scl={scl}')

        super().__init__(
            '&'.join(components),
            regex=True
        )

    def warp_coordinates(self, coords):
        return wgs2gcj(*coords)



import numpy as np


def weighted_percentile(a, weights, p):
    from scipy.interpolate import interp1d

    weighted = [(x, w) for x, w in zip(a, weights)]
    weighted.sort(key=lambda xw: xw[0])
    xs, ws = np.asarray(weighted).T
    ws /= ws.sum() / 100
    ps = np.r_[0, ws.cumsum()]
    ps = ps[:-1] + np.diff(ps) / 2

    interp = interp1d(ps, xs, fill_value='extrapolate')

    return interp(p)



import numpy as np

from metalpy.scab.utils.misc import Field
from metalpy.utils.taichi import ti_FieldsBuilder
from metalpy.utils.type import notify_package
from .kernel import kernel_matrix_forward
from .solver import DemagnetizationSolver


class IntegratedSolver(DemagnetizationSolver):
    def __init__(
            self,
            receiver_locations: np.ndarray,
            xn: np.ndarray,
            yn: np.ndarray,
            zn: np.ndarray,
            base_cell_sizes: np.ndarray,
            source_field: Field,
            is_cpu: bool
    ):
        """该函数直接计算核矩阵

        Parameters
        ----------
        receiver_locations
            观测点
        xn, yn, zn
            网格边界
        base_cell_sizes
            网格最小单元大小
        is_cpu
            是否是CPU架构，否则需要在对应设备上分配返回值矩阵

        Notes
        -----
        优势：

        - 简单直观

        缺陷：

        - 存在taichi的int32索引限制
        """
        super().__init__(receiver_locations, xn, yn, zn, base_cell_sizes, source_field)
        self.is_cpu = is_cpu

        nC = xn.shape[0]
        nObs = receiver_locations.shape[0]

        if is_cpu:
            self.builder = None
            self.A = np.empty((3 * nObs, 3 * nC), dtype=self.kernel_type)
        else:
            self.builder = builder = ti_FieldsBuilder()
            self.A = builder.place_dense((3 * nObs, 3 * nC), self.kernel_type)

    def build_kernel(self, model):
        # TODO: 实现进度条
        kernel_matrix_forward(
            self.receiver_locations,
            self.xn, self.yn, self.zn,
            self.base_cell_sizes, model,
            *[0] * 6, self.A,
            write_to_mat=True, compressed=False
        )

    def solve(self, magnetization):
        if self.is_cpu:
            Amat = self.A
        else:
            Amat = self.A.to_numpy()
        return solve_Ax_b(Amat, magnetization)


def solve_Ax_b(A, m):
    try:
        import pyamg
        x, _ = pyamg.krylov.bicgstab(A, m)
    except ModuleNotFoundError:
        notify_package(
            pkg_name='pyamg',
            reason='`pyamg` not found, falling back to slower version `np.linalg.solve`.'
        )
        x = np.linalg.solve(A, m)

    return x



from metalpy.scab.modelling.shapes import Ellipsoid, Tunnel, Prism
from metalpy.scab.modelling.shapes.testing.shape3d_place_test import assert_placing_inside


def test_ellipsoid():
    assert_placing_inside(Ellipsoid(2, 3, 4))


def test_tunnel():
    assert_placing_inside(Tunnel([0, 0, 0], 3, 6, 3))


def test_prism():
    assert_placing_inside(Prism([
        [0, 0], [1, 1], [0, 2], [2, 2], [2, 0]
    ], 3, 4))



from __future__ import annotations

import copy
import functools
import inspect
import os
import warnings
from typing import Sized

import numpy as np
import taichi as ti
from taichi.lang.kernel_impl import _kernel_impl
from taichi.lang.util import to_taichi_type, to_numpy_type

from .file import make_cache_directory, make_cache_file

ti_cache_prefix = 'taichi_cache'

ti_default_args = {
    'arch': ti.cpu,
    'offline_cache': True,
    'offline_cache_file_path': make_cache_directory(ti_cache_prefix),
}
ti_args = copy.deepcopy(ti_default_args)
ti_inited = False

ti_size_t = ti.i32
ti_size_dtype = to_numpy_type(ti_size_t)
ti_size_max = np.iinfo(ti_size_dtype).max


def ti_prepare(**kwargs) -> bool:
    """修改用于初始化taichi上下文的配置参数

    Parameters
    ----------
    kwargs
        需要修改的参数，以关键字参数形式指定

    Returns
    -------
    conflicted
        指示新的配置是否与已有配置冲突
    """
    conflicts = False
    for k, v in kwargs.items():
        if k not in ti_args or ti_args[k] != v:
            conflicts = True
        ti_args[k] = v
    return conflicts


def ti_init_once():
    global ti_inited
    if not ti_inited:
        ti_reinit()


def ti_reinit():
    global ti_inited
    ti.init(**ti_args)
    ti_inited = True


def ti_reset(*, params=True, now=False):
    """重置taichi的配置参数，使得下次调用涉及kernel的函数时会重新初始化

    同时也用于解决taichi目前没有gc导致内存泄露的问题
    taichi-dev/taichi#6803: Improve garbage collection
    https://github.com/taichi-dev/taichi/issues/6803

    Notes
    -----
    reset会在下次 `ti_init_once` 时生效，因此不会立刻解决内存泄露的问题，可以调用 `ti_reinit` 立刻强制重新初始化
    """
    global ti_inited, ti_args
    if params:
        ti_args = copy.deepcopy(ti_default_args)
    ti_inited = False
    if now:
        ti_reinit()


def ti_reset_now(*, params=True):
    ti_reset(params=params, now=True)


class WrappedTaichiContext:
    def __init__(self, **kwargs):
        self.backup = ti_args.copy()
        self.kwargs = kwargs

    def __enter__(self):
        ti_prepare(**self.kwargs)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        ti_prepare(**self.backup)


def ti_config(**kwargs):
    return WrappedTaichiContext(**kwargs)


def ti_arch(arch):
    arch_type = type(ti.cpu)
    if isinstance(arch, str):
        return getattr(ti, arch)
    elif isinstance(arch, arch_type):
        return arch
    elif isinstance(arch, list) and isinstance(arch[0], arch_type):
        return arch
    else:
        raise ValueError(f"Unknown architecture, "
                         f"should be one of 'cpu', 'gpu' or specific backends like 'cuda' or 'vulkan'. ")


def ti_func(fn, is_real_function=None):
    if is_real_function is None:
        return ti.func(fn)
    else:
        return ti.func(fn, is_real_function=is_real_function)


def ti_real_func(fn):
    return ti_func(fn, is_real_function=True)


def ti_pyfunc(fn):
    return ti.pyfunc(fn)


def ti_kernel(fn):
    ti_wrapped = _kernel_impl(fn, level_of_class_stackframe=3)

    @functools.wraps(ti_wrapped)
    def auto_init_wrapped(*args, **kwargs):
        ti_init_once()  # 完成初始化
        return ti_wrapped(*args, **kwargs)

    return auto_init_wrapped


def ti_data_oriented(cls):
    return ti.data_oriented(cls)


def ti_ndarray(dtype, shape):
    ti_init_once()
    return ti.ndarray(dtype, shape)


def ti_zeros_like(arr, *, dtype=None):
    ti_init_once()
    if dtype is None:
        dtype = to_taichi_type(arr.dtype)
    return ti.ndarray(dtype, arr.shape)


def ti_root():
    ti_init_once()
    return ti.root


def ti_cfg():
    ti_init_once()
    return ti.cfg


def ti_field(dtype,
             shape=None,
             order=None,
             name="",
             offset=None,
             needs_grad=False,
             needs_dual=False):
    ti_init_once()
    if 'Type' in type(dtype).__name__ and type(dtype).__module__.startswith('taichi'):
        # 有可能是quant类型，此时不需要转换
        dt = dtype
    else:
        dt = to_taichi_type(dtype)

    return ti.field(dt, shape, order, name, offset, needs_grad, needs_dual)


class WrappedFieldsBuilder:
    def __init__(self):
        self.fields_builder = ti.FieldsBuilder()
        self.snode_tree = None

    def finalize(self, raise_warning=True):
        self.snode_tree = self.fields_builder.finalize(raise_warning)
        return self.snode_tree

    def destroy(self):
        if self.snode_tree is not None:
            self.snode_tree.destroy()

    def place_dense(self, shape, *arr_types, axes=None):
        if axes is None:
            if isinstance(shape, Sized):
                ndim = len(shape)
            else:
                ndim = 1
            axes = ti.axes(*list(range(ndim)))
        arr_types = tuple(wrap_type_as_ti_field(t) for t in arr_types)
        self.dense(axes, shape).place(*arr_types)

        if len(arr_types) == 1:
            return arr_types[0]
        else:
            return arr_types

    def place_dense_like(self, arr, *, dtype=None, axes=None):
        if dtype is None:
            dtype = arr.dtype

        return self.place_dense(arr.shape, to_taichi_type(dtype), axes=axes)

    def __getattr__(self, name):
        return getattr(self.fields_builder, name)

    def __enter__(self) -> 'WrappedFieldsBuilder' | ti.FieldsBuilder:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.destroy()


def wrap_type_as_ti_field(t):
    if not isinstance(t, ti.Field):
        return ti_field(t)
    return t


def ti_FieldsBuilder():
    ti_init_once()
    return WrappedFieldsBuilder()


def ti_test_snode_support() -> bool:
    """测试cuda版本是否支持SNode和FieldsBuilder

    如果不支持，则需要使用动态数组ti.ndarray来代替

    Returns
    -------
    ret
        当前显卡cuda版本是否支持SNode和FieldsBuilder
    """
    ti_init_once()
    snode_support_indicator_cache = make_cache_file(f'{ti_cache_prefix}/snode_support')
    if os.path.exists(snode_support_indicator_cache):
        with open(snode_support_indicator_cache, 'r') as f:
            return f.read() == '1'
    try:
        with ti_FieldsBuilder() as builder:
            dummy = ti_field(ti.f64)
            builder.dense(ti.ij, (1, 1)).place(dummy)
            builder.finalize()
            _ = dummy.to_numpy()
            ret = True
    except RuntimeError as e:
        ret = False

    with open(snode_support_indicator_cache, 'w') as f:
        f.write('1' if ret else '0')

    return ret


def ti_ndarray_from(arr, sdim=0, dtype=None, field=False):
    """从其他数组类型创建Taichi ndarray

    Parameters
    ----------
    arr
        外部数组
    sdim
        结构维度，0代表Scalar，1代表Vector，2代表Matrix
    dtype
        指定taichi数组的数据类型，支持使用taichi或numpy的dtype标识
    field
        指示是否构建ti.field实例，若为False则构建ti.ndarray实例

    Returns
    -------
    ret
        返回结构维度为sdim的Taichi ndarray，且内容与arr相同
    """
    container = ti_ndarray_like(arr, sdim=sdim, dtype=dtype, field=field)
    copy_from(container, arr)
    return container


def ti_ndarray_like(arr, sdim=0, dtype=None, field=False, builder: ti.FieldsBuilder = None):
    """从其他数组类型创建相同尺寸的Taichi ndarray

    Parameters
    ----------
    arr
        外部数组
    sdim
        结构维度，0代表Scalar，1代表Vector，2代表Matrix
    dtype
        指定taichi数组的数据类型，支持使用taichi或numpy的dtype标识
    field
        指示是否构建ti.field实例，若为False则构建ti.ndarray实例
    builder
        指定用于分配数据的builder

    Returns
    -------
    ret
        返回结构维度为sdim的Taichi ndarray

    Notes
    -----
    field和ndarray区别参见：https://docs.taichi-lang.cn/docs/ndarray/
    """
    ti_init_once()

    if dtype is None:
        dtype = arr.dtype

    dt = to_taichi_type(dtype)
    if sdim == 0:
        shape = arr.shape
    elif sdim == 1:
        dt = ti.types.vector(arr.shape[-1], dt)
        shape = arr.shape[:-1]
    elif sdim == 2:
        dt = ti.types.matrix(*arr.shape[-sdim:], dt)
        shape = arr.shape[:-sdim]
    else:
        raise ValueError(f"Unsupported structure dim: {sdim}")

    if builder is None:
        if field:
            container = ti.field(dtype=dt, shape=shape)
        else:
            container = ti.ndarray(dtype=dt, shape=shape)
    else:
        container = ti.field(dt)
        builder.dense(ti.axes(*range(len(shape))), shape).place(container)

    return container


def copy_from(ti_arr, ext_arr):
    typeinfo = type(ext_arr)
    typename = typeinfo.__name__
    if typename == 'ndarray':
        ti_arr.from_numpy(ext_arr)
    else:
        raise ValueError(f"Unsupported array type: {typename}")
    # taichi暂未为ndarray实现这些来源
    # elif typeinfo.__module__ == 'torch':
    #     container.from_torch(arr)
    # elif typeinfo.__module__ == 'paddle':
    #     container.from_paddle(arr)


def ti_get_raw_function(kernel):
    raw = getattr(kernel, 'func', None) or getattr(kernel, '_primal', None) or kernel
    raw = getattr(raw, 'func', None) or raw

    return raw


def wrappable_for_ti_func(annotator):
    @functools.wraps(annotator)
    def wrapper(actual_func):
        # 用于处理ti.func或ti.kernel包装的函数
        # 这两类包装主要将实际功能函数添加到函数对象的__dict__中供调用
        funcs = []
        if getattr(actual_func, '_is_taichi_function', False):
            funcs.extend(['func'])
        elif getattr(actual_func, '_is_wrapped_kernel', False):
            funcs.extend(['_primal', '_adjoint'])

        ret = annotator(actual_func)

        for k in funcs:
            setattr(ret, k, annotator(getattr(ret, k)))

        return ret

    return wrapper


def check_contiguous(arr, name=None):
    typename = type(arr).__name__
    if typename == 'ndarray':
        import numpy as np
        if not arr.flags.c_contiguous and not arr.flags.f_contiguous:
            if name is None:
                loc = ''
            else:
                loc = f' at `{name}`'
            warnings.warn(f'Uncontiguous array detected{loc}.'
                          f' Copying to contiguous memory,'
                          f' which will consume extra memory.',
                          stacklevel=3)
            return np.ascontiguousarray(arr)
    else:
        # TODO: Torch、Paddle之类也可以检查？
        pass

    return arr


@wrappable_for_ti_func
def ensure_contiguous(func):
    """用于保证输入数组为连续数组（Taichi要求kernel的输入为连续数组）

    Parameters
    ----------
    func
        待包装的kernel函数

    Returns
    -------
    ret
        会自动查询输入的参数是否为连续数组，若否则拷贝一份到连续内存上

    Notes
    -----
    可能会导致引用数组地址变化，如果用于接受输出的数组被转换，则可能会导致输出到错误的位置
    """
    specs = inspect.getfullargspec(ti_get_raw_function(func))

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        n_args = len(args)
        n_varargs = len(specs.args) - n_args
        arg_names = specs.args + ([f'__vararg{i}' for i in range(-n_varargs)] if n_varargs < 0 else [])

        args = tuple(check_contiguous(arg, name=name) for arg, name in zip(args, arg_names))
        kwargs = {k: check_contiguous(v, name=k) for k, v in kwargs.items()}

        func(*args, **kwargs)

    return wrapper



import functools
import warnings

from metalpy.utils.string import format_string_list


class Mixin:
    def __init__(self, this):
        """定义一个Mixin类，用于在其它类中添加新的功能
        接受MixinManager的管辖，MixinManager负责管理Mixin的生命周期与信息交换

        Note
        ----
            1. 通过MixinManager的add方法将Mixin添加到目标类中（通常依靠Patch来通知PatchContext执行）
            2. MixinManager会将其中所有方法的第二个参数绑定到目标类上，此时在目标类和mixin类中调用的方法都是已绑定了两个类的方法，形如

                def xxx(self, this, ...)，其中self是mixin对象实例，this是对应的目标类实例

            3. 可以通过MixinManager的get方法获取Mixin，例如

                mixinA = mixed_class_instance.mixins.get(MixinA)

            同时一些特殊方法会被排除: 1. 私有方法（即名字以__开头） 2. 魔术方法（即名字被__包围） 3. 属性方法
        """
        pass

    def post_apply(self, this):
        pass


def before(func=None, *, target=None):
    return TaggedMethod(func, tag=TaggedMethod.Before, target=target)


def after(func=None, *, target=None):
    return TaggedMethod(func, tag=TaggedMethod.After, target=target)


def replaces(func=None, *, keep_orig=False, target=None):
    return TaggedMethod(func, tag=TaggedMethod.Replaces, keep_orig=keep_orig, target=target)


class TaggedMethod:
    Before = 'before'
    After = 'after'
    Replaces = 'replaces'
    All = {Before, After, Replaces}

    def __init__(self, func, tag, target=None, keep_orig=False):
        self.func = func
        self.tag = tag
        self.target = target
        self.keep_orig = keep_orig

    def __new__(cls, func, *args, **kwargs):
        if func is not None:
            return TaggedMethod(*args, **kwargs)(func)
        else:
            return super().__new__(cls)

    def __call__(self, func):
        self.func = func

        @functools.wraps(self.func)
        def wrapper(*args, **kwargs):
            return self.func(*args, **kwargs)

        wrapper.__tags__ = self
        return wrapper

    @staticmethod
    def verify(tag):
        if tag not in TaggedMethod.All:
            warnings.warn(f'Unknown mixin method type {tag},'
                          f' must be one of {format_string_list(TaggedMethod.All)}.'
                          f' Defaults to `{TaggedMethod.Replaces}`.')
            return TaggedMethod.Replaces
        return tag

    @staticmethod
    def check(meth):
        return getattr(meth, '__tags__', None)



class UnFinished:
    pass


_unfinished = UnFinished()


class LazyEvaluator:
    def __init__(self, func, *args, **kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.rval = _unfinished

    def get(self):
        if isinstance(self.rval, UnFinished):
            self.rval = self.func(*self.args, **self.kwargs)
        return self.rval

    def result(self):
        return self.get()

    def __getitem__(self, key):
        return self.kwargs[key]

    def __setitem__(self, key, value):
        self.kwargs[key] = value



import traceback


def traverse_args(args, kwargs, func):
    # 对所有的参数进行处理
    _args = []
    for arg in args:
        _args.append(func(arg))

    _kwargs = {}
    for k, v in kwargs.items():
        _kwargs[k] = func(v)

    return _args, _kwargs


def structured_traverse(struct, func):
    if isinstance(struct, dict):
        ret = {key: structured_traverse(element, func) for key, element in struct.items()}
    elif isinstance(struct, list):
        ret = [structured_traverse(element, func) for element in struct]
    else:
        ret = func(struct)

    return ret


def exception_caught(func, *args, **kwargs):
    """用于处理loky无法正确识别并捕获异常的问题
    TODO: 参考multiprocessing进行重新创建堆栈，从而帮助调试
    """
    try:
        return func(*args, **kwargs)
    except BaseException as e:
        traceback.print_exc()
        raise e



from __future__ import annotations

import os
import warnings

from SimPEG.simulation import BaseSimulation

from metalpy.mexin import Mixin
from metalpy.mexin import Patch
from metalpy.utils.type import get_params_dict
from metalpy.utils.object_path import get_full_qualified_path
from metalpy.mexin.utils import TypeMap
from metalpy.scab.distributed.policies import Distributable

from .taichi_kernel_base import Profiler


class TaichiContext(Mixin):
    _implementations = TypeMap()

    def __init__(self, this, arch=None, max_cpu_threads=None, profile: Profiler | bool = False, **kwargs):
        """初始化taichi上下文

        Parameters
        ----------
        this
            Simulation类指针，由mixin manager传递
        arch
            taichi架构，例如ti.cpu或ti.gpu，默认ti.cpu
        max_cpu_threads
            cpu arch下的最大线程数，taichi默认为cpu核心数
        """
        from metalpy.utils.taichi import ti_prepare, ti_arch, ti_reset

        super().__init__(this)

        params = {}

        if arch is not None:
            params['arch'] = ti_arch(arch)

        profile = Profiler.of(profile)
        self.profile = profile
        if profile.needs_kernel_profiler:
            params['kernel_profiler'] = True

        if max_cpu_threads is not None:
            if max_cpu_threads < 0:
                # 需要注意taichi目前使用的为虚拟核心数 (taichi/program/compile_config.cpp)
                # cpu_max_num_threads = std::thread::hardware_concurrency();
                # 因此这里使用python自带的cpu_count()，而不是psutil的物理核心数，和taichi保证一致
                # 考虑到超线程的存在，-1可能仍然会导致计算机满载
                max_cpu_threads = os.cpu_count() + max_cpu_threads
            params['cpu_max_num_threads'] = max_cpu_threads

        if ti_prepare(**params, **kwargs):
            ti_reset(params=False)

    def post_apply(self, this):
        impl = TaichiContext._implementations.get(type(this))

        if impl is None:
            warnings.warn(f'Taichi support for {get_full_qualified_path(this)} is not implemented. Ignoring it.')
            return

        this.mixins.add(impl, profile=self.profile)


class Tied(Patch, Distributable):
    def __init__(self, arch=None, max_cpu_threads=None, **kwargs):
        super().__init__()
        self.params = get_params_dict(arch=arch, max_cpu_threads=max_cpu_threads, **kwargs)

    def apply(self):
        self.add_mixin(BaseSimulation, TaichiContext, **self.params)


def __implements(target):
    def decorator(func):
        TaichiContext._implementations.map(target, func)
        return func
    return decorator


@__implements('SimPEG.potential_fields.magnetics.simulation.Simulation3DIntegral')
def _():
    from .potential_fields.magnetics.simulation import TiedSimulation3DIntegralMixin
    return TiedSimulation3DIntegralMixin



from . import Coefficient


class PBR(Coefficient):
    def __init__(self, value):
        """目前 VTK 中的PBR相关参数都是以系数形式提供，因此直接继承自 `Coefficient`

        TODO: 将来如果有新特性可能需要跟进
        """
        super().__init__(value)

    def build(self):
        return {
            'pbr': True,
            **super().build()
        }


class Metallic(PBR):
    pass


class Roughness(PBR):
    pass



import numpy as np

from .array_ops import OperatorDispatcher, ArrayType

GetColumnsExcept = OperatorDispatcher('GetColumnsExcept')


def get_columns_except(arr, ind):
    """从数组获取除去指定列外的部分

    Parameters
    ----------
    arr
        数组实例
    ind
        列索引或列索引数组 / 切片slice / 列名或列名数组（pandas）

    Returns
    -------
    ret
        返回不在ind中指定的列
    """
    return GetColumnsExcept.dispatch(type(arr))(arr, ind)


def _excluded_mask(length, excludes):
    take = np.ones(length, dtype=bool)
    take[excludes] = False

    return take


@GetColumnsExcept.register(ArrayType.numpy)
def _get_columns_except_np(arr, ind):
    return arr[:, _excluded_mask(arr.shape[1], ind)]


@GetColumnsExcept.register(ArrayType.pandas)
def _get_columns_except_pd(arr, ind):
    if not isinstance(ind, slice):  # slice直接索引会默认索引行，需要排除
        try:
            _ = arr[ind]  # 判断是否为合法的列key
            keys = set(arr.keys())
            keys.difference_update(ind)
            return arr[list(keys)]
        except KeyError:
            pass

    return arr.iloc[:, _excluded_mask(arr.shape[1], ind)]  # 否则认为是列索引



from .executor import Executor, check_workers
from .queue_like_worker_context import QueueLikeWorkerContext
from .worker import Worker


class DaskExecutor(Executor):
    def __init__(self, scheduler_addr, extra_paths=None, excludes=None, upload_modules=True):
        """构造基于dask.distributed的执行器

        Parameters
        ----------
        scheduler_addr
            调度器地址，包括协议和端口号，例如 "tcp://xx.yy.zz:8786"
        extra_paths
            额外需要自动搜索并上传代码到集群的模块路径
        excludes
            需要排除不自动搜索并上传代码到集群的模块路径
        upload_modules
            指定是否自动搜索代码并上传到集群

        Notes
        -----
        在启用自动上传模式时，有时会导致worker中存在多个相同的类，
        形似 `Scene.of` 一样使用 `isinstance` 的语句有时会出现无法识别的问题，
        """
        import uuid
        from distributed import Client
        from .dask_helper import configure_dask_client

        super().__init__()
        self.client = Client(scheduler_addr)
        if upload_modules:
            configure_dask_client(self.client, extra_paths=extra_paths, excludes=excludes)

        worker_groups = {}
        for addr, worker in self.client.scheduler_info()['workers'].items():
            host = worker['host']
            if host not in worker_groups:
                worker_groups[host] = []
            # LocalCluster的名字有可能是数字，因此需要转换成字符串
            worker_groups[host].append(str(worker['name']))

        self.workers = []
        n_actual_working_units = 0
        load_weight = {}

        for host, workers in worker_groups.items():
            for worker in workers:
                worker_instance = Worker(worker, weight=1, verbose=False)
                worker_name = worker_instance.name

                if worker_name in load_weight:
                    weight = load_weight[worker_name]
                else:
                    weight = 1

                worker_instance.weight = weight
                self.workers.append(worker_instance)
                n_actual_working_units += weight

        self.client_id = str(uuid.uuid4())
        self._sub = None

    @property
    def sub(self):
        import distributed

        if self._sub is None:
            self._sub = distributed.Sub(self.client_id)

        return self._sub

    @property
    def has_sub(self):
        return self._sub is not None

    def map(self, func, *iterables, worker=None, workers=None, chunksize=None):
        kwargs = _extract_workers(None, workers)
        return self.gather(self.client.map(func, *iterables, **kwargs, batch_size=chunksize))

    def do_submit(self, func, *args, workers=None, **kwargs):
        option_kwargs = _extract_workers(None, workers)
        option_kwargs.update(kwargs)

        return self.client.submit(func, *args, **option_kwargs)

    def get_workers(self):
        return self.workers

    def is_local(self):
        return False

    def _gather_single(self, future):
        return self.client.gather(future)

    def scatter(self, data):
        return self.client.scatter(data)

    def get_worker_context(self):
        return DaskWorkerContext(self.client_id)

    def receive_events(self, timeout):
        import asyncio

        while True:
            try:
                event, msg = self.sub.get(timeout)
                yield event, msg
            except asyncio.exceptions.TimeoutError:
                yield None, None

    def shutdown(self, wait=True):
        # TODO: 实现wait，参考 distributed.cfexecutor.ClientExecutor.shutdown
        self.client.shutdown()


class DaskWorkerContext(QueueLikeWorkerContext):
    def __init__(self, client_id):
        super().__init__()
        self.client_id = client_id
        self._pub = None

    @property
    def queue(self):
        import distributed

        if self._pub is None:
            self._pub = distributed.Pub(self.client_id)

        return self._pub

    @queue.setter
    def queue(self, _):
        pass


def _extract_workers(worker, workers):
    workers = check_workers(worker, workers)

    kwargs = {}
    if workers is not None:
        targets = [worker.get_name() for worker in workers]
        kwargs['workers'] = targets
        kwargs['allow_other_workers'] = True

    return kwargs



from typing import Iterable

import numpy as np

from metalpy.scab.utils.misc import Field
from .utils import get_demag_apparent_k


class FactoredDemagnetization:
    def __init__(self, n, source_field=None):
        """
        依据退磁因子，计算退磁作用下的磁化强度，通过退磁因子计算视磁化率

        Parameters
        ----------
        n
            三轴退磁系数
        source_field
            外部场源，也可以在dpred时指定，主要为保持和Simulation的一致性

        Notes
        -----
        当某一方向系数为None时自动推断，所有None值假设为相同的值，且值为 (1 - sum(n[n != None])) / len(n[n == None])
        """
        super().__init__()
        if isinstance(n, Iterable):
            n = np.asarray(n)
            if np.any(n == None):
                n[n == None] = (1 - sum(n[n != None])) / len(n[n == None])
                n = n.astype(np.float64)
        self.N = n
        self.source_field = source_field

    def dpred(self, model, source_field: Field = None):
        """
        Parameters
        ----------
        model
            array-like(nC,)，磁化率模型
        source_field
            外部场源，若为空，则返回视磁化率矩阵，否则返回三轴等效磁化率

        Returns
        -------
        ret
            array(nC, 3)，
            若source_field为空，返回三轴视磁化率；
            若source_field非空，返回在该场作用下的三轴等效磁化率（三轴磁化强度除以场源强度）
        """
        if source_field is None:
            source_field = self.source_field

        model = np.asarray(model)
        k = get_demag_apparent_k(model, self.N)

        if source_field is not None:
            k = k * Field(source_field).unit_vector

        return k



from metalpy.scab.modelling.shapes import BarFramework, Ellipsoid
from metalpy.scab.modelling.shapes.testing.shape3d_place_test import assert_placing_inside


def shp():
    return Ellipsoid(10, 10, 10)


def _test_bar_framework(inherit_transform):
    kwargs = {
        'spec': 1,
        'n_rooms': 2,
        'inherit_transform': inherit_transform
    }
    assert_placing_inside(BarFramework(shp().rotated(0, 0, 45).translated(10, 10, 10), **kwargs))
    assert_placing_inside(BarFramework(shp(), **kwargs).rotated(0, 0, 45).translated(10, 10, 10))
    assert_placing_inside(BarFramework(shp().rotated(0, 0, 45), **kwargs).translate(10, 10, 10))


def test_bar_framework():
    _test_bar_framework(True)


def test_bar_framework_inherit_transform():
    _test_bar_framework(False)



import inspect
import types
import warnings

from metalpy.mexin.injectors.replacement import get_orig


def wrap_method_with_target(target, func):
    if target is not None and not isinstance(target, (type, type(types))):
        # 目标不是None、类型与模块，则是实例，func是实例方法，直接绑定
        wrapper = types.MethodType(func, target)
        is_target_method = True
    else:
        wrapper = func
        is_target_method = False

    return wrapper, is_target_method


def update_params(func, args, kwargs, new_kwargs):
    """使用new_kwargs更新func的args和kwargs

    Parameters
    ----------
    func
        目标函数签名
    args
        目标函数原有的位置参数
    kwargs
        目标函数原有的关键字参数
    new_kwargs
        新的参数列表，使用str修改关键字或非varargs的位置参数，使用int修改varargs中指定位置的参数

    Returns
    -------
    updated_args, updated_kwargs
        更新后的args和kwargs
    """
    from metalpy.utils.type import not_none_or_default

    # func可能被替换过，导致函数签名和原函数不一致，因此尝试回溯原函数从而获取真实函数签名
    # get_orig在不存在原函数时会返回None，因此需要判断
    func = not_none_or_default(get_orig(func), _default=func)
    arg_spec = inspect.getfullargspec(func)
    updated_args = None
    updated_kwargs = None
    new_kwargs = dict(new_kwargs)

    arg_spec_args = arg_spec.args
    if isinstance(func, types.MethodType):
        # func是实例方法，需要排除掉self参数
        arg_spec_args = arg_spec_args[1:]

    for i, arg_name in enumerate(arg_spec_args):
        new_arg = new_kwargs.pop(arg_name, None)
        if new_arg is not None:
            if i < len(args):
                # 如果下标在args范围内，则一定是对应位置的参数
                if updated_args is None:
                    updated_args = list(args)
                updated_args[i] = new_arg
            else:
                # 如果下标不在args范围内，则原参数没有被指定或在kwargs中指定，无论哪种情况都应放到kwargs中
                if updated_kwargs is None:
                    updated_kwargs = dict(kwargs)
                updated_kwargs[arg_name] = new_arg

    for i in range(len(args)):
        new_arg = new_kwargs.pop(i, None)
        if new_arg is not None:
            # 通过下标替换args中的参数，主要用于varargs参数
            if updated_args is None:
                updated_args = list(args)
            updated_args[i] = new_arg

    for arg_name in arg_spec.kwonlyargs:
        new_arg = new_kwargs.pop(arg_name, None)
        if new_arg is not None:
            # 通过关键字替换kwargs中的kwonlyargs
            if updated_kwargs is None:
                updated_kwargs = dict(kwargs)
            updated_kwargs[arg_name] = new_arg

    if arg_spec.varkw is not None:
        # 其他参数都已经放入对应位置
        # 有varkw，则将剩下的关键字参数全部传入
        if updated_kwargs is None:
            updated_kwargs = dict(kwargs)
        for k in list(new_kwargs.keys()):
            if not isinstance(k, str):
                # new_kwargs里还有数字这类位置参数或者甚至其他错误参数
                warnings.warn(f'update_params received invalid parameter key: {k}. Ignoring it.')
                new_kwargs.pop(k)
        updated_kwargs.update(new_kwargs)

    if updated_args is None:
        updated_args = args
    if updated_kwargs is None:
        updated_kwargs = kwargs

    return updated_args, updated_kwargs



from __future__ import annotations

import PIL.Image
import numpy as np
from PIL.Image import Image

from metalpy.carto.coords.coordinates import CRSQuery
from metalpy.carto.utils.crs import CRSLike
from metalpy.carto.pyvista.crs import warp_dataset
from metalpy.utils.bounds import Bounds

from .geo_image_ref_system import GeoImageRefSystem


class GeoImage:
    def __init__(self,
                 image: Image,
                 offset,
                 ref_system: GeoImageRefSystem):
        """包含地理坐标信息的图像

        Parameters
        ----------
        image
            原始图像
        offset
            图像原点到坐标系原点的单元距离，图像原点定义在上或在下与`ref_system.flip_y`有关
        ref_system
            参考坐标系

        Notes
        -----
        当`flip_y=True`，地理坐标系和图像坐标系的y轴方向相反，如图所示
        此时offset为从坐标系原点到图像左下角的距离
        ↑ unit_y (m)
        │  ┌────────────────────────→ img_x (px)
        │  │┌────────────────────┐
        │  ││                    │
        │  ││          ┌─────┐   │
        │  ││          │     │   │
        │  ││          └─────┘   │
        │  │└────────────────────┘
        │  ↓ img_y (px)
        └───────────────────────────────────→ unit_x (m)

        当`flip_y=False`，地理坐标系和图像坐标系的y轴方向相同，如图所示
        此时offset为从坐标系原点到图像左上角的距离
        ↑ unit_y (m)
        │  ↑ img_y (px)
        │  │┌────────────────────┐
        │  ││                    │
        │  ││          ┌─────┐   │
        │  ││          │     │   │
        │  ││          └─────┘   │
        │  │└────────────────────┘
        │  └────────────────────────→ img_x (px)
        └───────────────────────────────────→ unit_x (m)
        """
        self.image = image
        self.offset = np.array(offset)
        self.ref_system = ref_system

    @property
    def unit_size(self):
        return self.ref_system.unit_size

    @property
    def size(self):
        return self.image.size

    @property
    def width(self):
        return self.size[0]

    @property
    def height(self):
        return self.size[1]

    @property
    def bounds(self):
        return np.c_[
            self.offset,
            self.offset + self.size
        ].ravel().view(Bounds)

    @property
    def geo_bounds(self):
        """边界像素的中心点所形成的边界框
        """
        return self.ref_system.to_geo_bounds(
            self.bounds
            + (0.5, -0.5, 0.5, -0.5)
        )

    @property
    def edge_geo_bounds(self):
        """边界像素的外边界框
        """
        return self.ref_system.to_geo_bounds(
            self.bounds
        )

    @property
    def flip_y(self):
        return self.ref_system.flip_y

    @property
    def crs(self):
        return self.ref_system.crs

    @property
    def mode(self):
        return self.image.mode

    def crop(self, bounds=None, geo_bounds=None):
        if bounds is None:
            assert geo_bounds is not None, 'Either one of `bounds` or `geo_bounds` is required.'
            bounds = self.ref_system.to_unit_bounds(geo_bounds)

        corners = bounds.as_corners()
        offset = corners.origin
        rel_corners = corners - self.bounds.as_corners().origin
        if self.flip_y:
            rel_corners.origin[1], rel_corners.end[1] = (
                self.height - rel_corners.end[1],
                self.height - rel_corners.origin[1]
            )

        return GeoImage(
            image=self.image.crop(rel_corners.ravel()),
            offset=offset,
            ref_system=self.ref_system
        )

    def paste(self, img: 'GeoImage'):
        """将新地理图像`img`粘贴到当前图像（会修改当前图像）

        Parameters
        ----------
        img
            待粘贴的地理图像
        """
        if self.ref_system.flip_y:
            offset = (
                img.bounds[0] - self.bounds[0],
                self.bounds[3] - img.bounds[3]
            )
        else:
            offset = self.bounds[[0, 2]] - img.bounds[[0, 2]]

        self.image.paste(img.image, offset)

    def to_polydata(self,
                    dest_crs: CRSLike | None = None,
                    query_dest_crs: CRSQuery | None = None,
                    return_crs=False
                    ):
        """将GeoImage实例转换为PyVista对象

        Parameters
        ----------
        dest_crs
            需要转换到的目标坐标系
        query_dest_crs
            需要转换到的目标坐标系（但通过查询获取），用于搜索UTM等分区坐标系
        return_crs
            指示是否返回转换后得到的坐标系

        Returns
        -------
        image_obj
            转换后得到的PyVista对象，为UnstructuredGrid
        """
        import pyvista as pv

        bounds = self.edge_geo_bounds
        xs = np.linspace(bounds.xmin, bounds.xmax, self.width + 1)
        ys = np.linspace(bounds.ymin, bounds.ymax, self.height + 1)

        if self.flip_y:
            ys = ys[::-1]

        grid = pv.RectilinearGrid(xs, ys, 0)

        img = np.asarray(self.image)
        n_channels = img.shape[2] if img.ndim > 2 else 1
        image_data = img.reshape(-1, n_channels)

        data_name = 'Image'
        alias = None  # PyVista中绘制`_rgb`或`_rgba`后缀的数据时会自动启用彩色模式（rgb=True）
        if n_channels == 3:
            alias = f'{data_name}_rgb'
        elif n_channels == 4:
            alias = f'{data_name}_rgba'

        grid[data_name] = image_data
        if alias is not None:
            grid[alias] = image_data
            grid.set_active_scalars(alias)

        grid = grid.cast_to_unstructured_grid()

        if return_crs:
            result = (grid, self.crs)
        else:
            result = grid

        if dest_crs is not None or query_dest_crs is not None:
            result = warp_dataset(
                grid,
                src_crs=self.crs,
                crs=dest_crs,
                query=query_dest_crs,
                inplace=True,
                return_crs=return_crs
            )

        return result

    def save(self, fp, format=None, **params):
        self.image.save(fp, format=format, **params)

    def save_geo_tiff(self, path):
        import rasterio
        from rasterio.transform import Affine

        # 需要注意图像矩阵存储向下为正，而数据向上（北）为正。
        # offset为左上角像素中心点位置，左上角实际为(xmin, ymax)，
        # 且y_scale < 0，即图像行索引增加方向为坐标值减少方向（南方向）
        bounds = self.geo_bounds
        transform = Affine.translation(bounds.xmin, bounds.ymax) * Affine.scale(self.unit_size[0], -self.unit_size[1])
        image_arr = np.asarray(self.image)

        if image_arr.ndim < 3:
            channels = 1
        else:
            channels = image_arr.shape[2]

        with rasterio.open(
            path,
            'w',
            driver='GTiff',
            width=self.width,
            height=self.height,
            count=channels,
            dtype=image_arr.dtype,
            crs=self.crs,
            transform=transform,
        ) as dataset:
            if image_arr.ndim < 3:
                dataset.write(image_arr, 1)
            else:
                for dim in range(channels):
                    dataset.write(image_arr[:, :, dim], dim + 1)

    @staticmethod
    def read(path):
        import rasterio
        from rasterio.plot import reshape_as_image
        from affine import Affine

        with rasterio.open(path) as dataset:
            transform = dataset.transform
            assert transform.identity() == Affine(1.0, 0.0, 0.0, 0.0, 1.0, 0.0), \
                'GeoImage supports only translation and scaling as transform.'

            origin = (transform.xoff, transform.yoff)
            unit_size = (transform[0], transform[4])
            ref = GeoImageRefSystem.of_unit_center_bounds(
                center_origin=origin,
                unit_size=unit_size,
                crs=dataset.crs
            )
            raster = reshape_as_image(dataset.read())
            image = PIL.Image.fromarray(raster)

            if not ref.flip_y:
                offset = (0, 0)
            else:
                offset = (0, -image.height)

            return ref.map_image(image, offset=offset)



import numpy as np

from metalpy.utils.dhash import dhash
from metalpy.utils.bounds import Bounds
from . import Shape3D


def is_abs_distance_in(arr, x0, r):
    return np.abs(arr - x0) < r


class Tunnel(Shape3D):
    def __init__(self, p0, r0, r1, L):
        """定义一个起始点在p0，内外圆半径r0和r1，向x方向延申长度为L的空心圆柱环

        Parameters
        ----------
        p0
            起始点，隧道的左中位置
        r0
            内圆半径
        r1
            外圆半径
        L
            隧道长度
        """
        super().__init__()
        self.p0 = np.asarray(p0)
        self.r0 = r0
        self.r1 = r1
        self.L = L

    def do_place(self, mesh_cell_centers, progress):
        x0, y0, z0 = self.p0
        L = self.L
        r0, r1 = self.r0, self.r1
        xs, ys, zs = mesh_cell_centers[:, 0], mesh_cell_centers[:, 1], mesh_cell_centers[:, 2]
        # 判定是否在长L，宽高r1的长方体范围内
        included = \
            is_abs_distance_in(xs, x0 + L / 2, L / 2) & \
            is_abs_distance_in(ys, y0, r1) & \
            is_abs_distance_in(zs, z0, r1)

        indices = included

        possible_grids = mesh_cell_centers[included, :]  # 只计算在这个长方体区域内的方块是否属于圆环柱
        rs = np.sqrt((possible_grids[:, 1] - self.p0[1]) ** 2 + (possible_grids[:, 2] - self.p0[2]) ** 2)
        indices[included] = (rs < r1) & (rs > r0)

        return indices

    def __dhash__(self):
        return dhash(super().__dhash__(),
                     self.r0, self.r1, self.L, *self.p0)

    def do_clone(self, deep=True):
        return Tunnel(
            self.p0.copy(),
            self.r0,
            self.r1,
            self.L
        )

    @property
    def local_bounds(self):
        x0, y0, z0 = self.p0
        r1 = self.r1
        L = self.L
        return Bounds(np.r_[
            x0, L,
            y0 - r1, y0 + r1,
            z0 - r1, z0 + r1,
        ].ravel())

    def to_local_polydata(self):
        import pyvista as pv

        resolution = 100
        a = np.linspace(0, 2 * np.pi, resolution + 1)[:-1]  # 2π位置和0位置重复，去掉最后一个
        x0, y0, z0 = self.p0
        r0, r1 = self.r0, self.r1
        L = self.L
        x = np.r_[x0, x0 + L]
        y = np.r_[y0 + r0 * np.cos(a), y0 + r1 * np.cos(a)]
        z = np.r_[z0 + r0 * np.sin(a), z0 + r1 * np.sin(a)]

        # array(resolution, ) [<x0内圆...>, <x0外圆...>, <x0+L内圆...>, <x0+L外圆...>]
        xs = x.repeat(2 * resolution)
        ys = np.tile(y, 2)
        zs = np.tile(z, 2)

        indices = np.arange(xs.shape[0]).reshape([4, -1])
        rolled = np.roll(indices, -1, axis=1)
        edge_counts = np.full(resolution, 3)

        face_vertices = [
            [0, 1],  # 底面
            [2, 3],  # 顶面
            [1, 3],  # 外圈表面
            [0, 2],  # 内圈表面
        ]

        faces = []
        for i0, i1 in face_vertices:
            faces.extend([  # 这个三角面
                edge_counts,
                indices[i0],
                indices[i1],
                rolled[i1]
            ])
            faces.extend([  # 那个三角面
                edge_counts,
                rolled[i1],
                rolled[i0],
                indices[i0]
            ])

        # 每次都 column_stac 然后 vstack 再 ravel
        # 和
        # 直接全部 column_stac 到一起再 ravel
        # 对PyVista而言等价
        faces = np.column_stack(faces).ravel()

        ret = pv.PolyData(np.c_[xs, ys, zs], faces=faces)
        ret.flip_normals()

        return ret

    @property
    def bottom_area(self):
        return np.pi * (self.r1 ** 2 - self.r0 ** 2)

    @property
    def volume(self):
        return self.bottom_area * self.L

    @property
    def area(self):
        ba = self.bottom_area
        p0 = 2 * np.pi * self.r0
        p1 = 2 * np.pi * self.r1

        return 2 * ba + (p0 + p1) * self.L



import contextlib
import re
import subprocess

from setuptools import setup
from pathlib import Path

from setuptools.command.sdist import sdist
from versioningit import get_cmdclasses


class RebuildSdistCommand(sdist):
    def exclude_untracked_source_files(self):
        """移除git未追踪的.py源文件
        TODO: 如果将来加入预生成的.py文件可能需要修改该函数
        """
        tracked_files = subprocess.check_output(['git', 'ls-files'], **{'text': 'utf-8'})
        tracked_files = tracked_files.strip().replace('\\', '/')  # 统一为反斜杠
        tracked_files = set(tracked_files.splitlines())

        files = self.filelist.files
        for i in range(len(files) - 1, 0, -1):
            if files[i].endswith('.py') and files[i].replace('\\', '/') not in tracked_files:
                print(f"removing untracked source file '{files[i]}'")
                del files[i]

    def rebuild_sdist(self):
        """因未知原因，打包出的sdist中路径会记录成绝对路径，因此在打包后添加一个workaround将路径转换为相对路径
        """
        gz_file = None
        for file in self.archive_files:
            if file.endswith('.tar.gz'):
                gz_file = Path(file)
                break

        if gz_file is not None:
            import gzip
            print(f'rebuilding {gz_file}')
            tar_file = gz_file.with_suffix('')
            g_file = gzip.GzipFile(gz_file)
            tar_file.write_bytes(g_file.read())
            g_file.close()

            with gzip.open(gz_file, "wb") as f_out:
                f_out.write(tar_file.read_bytes())

            tar_file.unlink()

    def make_distribution(self):
        self.exclude_untracked_source_files()

        super(RebuildSdistCommand, self).make_distribution()

        self.rebuild_sdist()


@contextlib.contextmanager
def generate_pypi_readme():
    """从仓库版README.md中生成适用于pypi的README文件

    Notes
    -----
        这是对当前setuptools系统不支持pyproject.toml联合使用setup.py动态添加long_description (readme) 的workaround
        该workaround的特殊处理包括：
            在MANIFEST.in中手动排除会被自动包含的README.md，并手动包含生成的文件README-pypi.md
            在pyproject.toml中指定readme = "README-pypi.md"

        注意事项：
            README.md中：
            所有.md或.rst结尾的引用会被替换为github地址的引用
            所有./都会被替换为github地址的raw对象地址引用
    """
    readme_file = Path("README.md")
    pypi_readme_file = Path('README-pypi.md')
    root = 'https://github.com/yanang007/metalpy'
    if not pypi_readme_file.exists():
        long_description = readme_file.read_text(encoding='utf-8')
        long_description = re.compile(r'\(\.?[/|\\]?(.*?\.(md|rst))\)') \
            .sub(rf'({root}/tree/main/\1)', long_description)
        long_description = long_description.replace('./', f'{root}/raw/main/')

        pypi_readme_file.write_text(long_description, encoding='utf-8')

    try:
        yield
    finally:
        if pypi_readme_file.exists() and readme_file.exists():
            # 如果两个README.md都存在，则应当清理掉临时生成的pypi-README.md
            # - 这种情况包含在项目路径下构建sdist的阶段
            # 否则如果只存在pypi-README.md，则不应该清理pypi-README.md
            # - 这种情况包含从sdist构建wheel的阶段，会创建一个只包含sdist中文件的临时目录，由于README.md被排除，所以此时不存在README.md
            pypi_readme_file.unlink()


with generate_pypi_readme():
    setup(
        cmdclass=get_cmdclasses({'sdist': RebuildSdistCommand}),
        # Other arguments go here
    )



from typing import Union

from .recoverable_injector import RecoverableInjector
from .utils import wrap_method_with_target
from .replacement import create_replacement, get_ancestor, get_nest


class Replaces(RecoverableInjector):
    def __init__(self, target, nest=None, keep_orig: Union[str, bool] = False, force_unbound=False):
        """替换目标函数

        Parameters
        ----------
        target
            待替换的目标，支持函数，方法，类名
        nest
            手动指定替换对象的所在空间，支持模块，类，实例
        keep_orig
            指示是否将原函数作为参数传入，当传入str时，将以该名字作为参数名
        force_unbound
            如果nest不为对象实例，该参数无效，如果nest为对象实例，则强制不对成员函数进行绑定

        Notes
        -----
            keep_orig=True时，原函数将以 **orig_fn** 的名字传入函数，此时请确保替换的函数包含这个名字的参数
            keep_orig为str时，原函数将以该名字传入函数，此时请确保替换的函数包含该指定名字的参数
        """
        super().__init__()
        target_name = target.__name__
        if nest is None:
            root_target = get_ancestor(target)
            nest = get_nest(root_target)
            if nest is None:
                raise NotImplementedError(
                    f'Unable locate where {root_target.__name__} is, consider specify it by "nest" param.'
                )

        self.nest = nest
        self.name = target_name
        self.backup = None
        self.keep_orig = keep_orig
        self.force_unbound = force_unbound

    def __call__(self, func):
        cmd = f'self.nest.{self.name}'
        self.backup = orig = eval(cmd)

        keep_orig = self.keep_orig
        if keep_orig and isinstance(keep_orig, bool):
            keep_orig = 'orig_fn'

        if keep_orig:
            orig_fn = {keep_orig: orig}
            wrapper = lambda *args, **kwargs: func(*args, **kwargs, **orig_fn)
        else:
            wrapper = lambda *args, **kwargs: func(*args, **kwargs)

        if not self.force_unbound:
            wrapper, is_method = wrap_method_with_target(self.nest, wrapper)
        wrapper = create_replacement(wrapper, orig, self)
        cmd = f'self.nest.{self.name} = wrapper'
        exec(cmd)

        return wrapper

    def rollback(self):
        cmd = f'self.nest.{self.name} = self.backup'
        exec(cmd)


def replaces(target, nest=None, keep_orig=False):
    """替换目标函数

    Parameters
    ----------
    target
        待替换的目标，支持函数，方法，类名
    nest
        手动指定替换对象的所在空间，支持模块，类，实例
    keep_orig
        指示是否将原函数作为参数传入，当传入str时，将以该名字作为参数名

    Notes
    -----
        keep_orig=True时，原函数将以 **orig_fn** 的名字传入函数，此时请确保替换的函数包含这个名字的参数
        keep_orig为str时，原函数将以该名字传入函数，此时请确保替换的函数包含该指定名字的参数
    """
    return Replaces(target, nest=nest, keep_orig=keep_orig)



from __future__ import annotations

import math
from pathlib import Path
from typing import Iterable, TYPE_CHECKING, Sequence

import numpy as np
from scipy.spatial.transform import Rotation

from metalpy.carto.coords import Coordinates
from metalpy.utils.file import PathLike
from metalpy.utils.matplotlib import check_axis

if TYPE_CHECKING:
    from matplotlib.colors import Colormap


class FlightPlanar:
    Left = Rotation.from_euler('z', 90, degrees=True)
    Right = Rotation.from_euler('z', -90, degrees=True)

    def __init__(self, sample_distance: float = 10):
        """实现类tutle接口的航线规划功能，并提供航线动态图绘制接口

        Parameters
        ----------
        sample_distance
            航线的采样点间距，以下所有规划的航线上按该航距采样位置点
        """
        self.paths = []
        self.position = np.asarray([0, 0])
        self._direction = np.asarray([0, 1])
        self.sample_distance = sample_distance

    @property
    def direction(self):
        return self._direction

    @direction.setter
    def direction(self, v):
        self._direction = v / np.linalg.norm(v)

    def add_path(self, path):
        self.paths.append(np.asarray(path))

    def rotate_direction(self, angle, radians=False):
        self.direction = FlightPlanar.rot2d(
            Rotation.from_euler('z', angle, degrees=not radians),
            self.direction
        )

    def forward(self, distances):
        n_steps = int(distances / self.sample_distance)
        origin = self.position
        end = origin + self.direction * self.sample_distance * n_steps
        xs = np.linspace(origin[0], end[0], n_steps + 1)
        ys = np.linspace(origin[1], end[1], n_steps + 1)
        self.add_path(np.c_[xs, ys])

        real_end = origin + self.direction * distances
        self.position = real_end

    def turn(self, radius, theta, radians=False):
        if not radians:
            theta = np.deg2rad(theta)
        rot = FlightPlanar.get_rot(-theta)
        ortho_dir = FlightPlanar.rot2d(rot, self.direction)

        center = self.position + ortho_dir * radius

        a0 = FlightPlanar.vec2rad(ortho_dir)
        a1 = a0 + theta
        a_step = self.sample_distance / radius * np.sign(theta)
        angles = np.r_[np.arange(a0, a1, a_step), a1]

        current_angle = FlightPlanar.vec2rad(self.direction)
        xs = radius * np.cos(angles)
        ys = radius * np.sin(angles)
        path = np.c_[xs, ys]
        path += center
        path += self.position - path[0]

        self.add_path(path[:-1])
        self.position = path[-1]
        self.direction = FlightPlanar.rad2vec(current_angle + theta)

    def stay(self, n_steps):
        path = self.position[np.newaxis, :].repeat(n_steps, axis=0)
        self.add_path(path)

    def spiral(self, r0, theta, round_increment, radians=False, return_radix=False):
        """等距螺旋线

        .. math::
            r(\\theta) = k\\theta + r_0

        Parameters
        ----------
        r0
            起始极径
        theta
            螺旋线旋转角（720°为2个周期，半径增加2 * `round_increment`）
        round_increment
            每周期极径增量 / 线距（k = `round_increment` / (2 * np.pi) ）
        radians
            传入值是否为弧度
        return_radix
            是否返回每一段的弧长，若为False则无返回值
        """
        if not radians:
            theta = np.deg2rad(theta)
        rot = FlightPlanar.get_rot(-theta)
        ortho_dir = FlightPlanar.rot2d(rot, self.direction)

        center = self.position + ortho_dir * r0
        a = 0
        inc = round_increment / np.pi / 2
        angles = []
        total = abs(theta)

        radix = None
        if return_radix:
            radix = []

        while a < total:
            radius_0 = r0 + a * inc
            const = FlightPlanar._spiral_arc_length(radius_0, inc) + self.sample_distance
            da = self.sample_distance / radius_0
            radius_x = radius_0 + da * inc

            # 牛顿法求解给定弧长下的等距螺旋线的极角增量  Rx+1 = Rx - F(Rx) / F'(Rx)
            radius_x = (
                    radius_x -
                    (FlightPlanar._spiral_arc_length(radius_x, inc) - const)
                    / FlightPlanar._spiral_arc_length_derive(radius_x, inc)
            )

            da = (radius_x - radius_0) / inc
            a = a + da
            angles.append(a)
            if radix is not None:
                radix.append((radius_0 + radius_x))

        angles = np.asarray(angles + [abs(theta)])
        radix = angles * inc + r0
        xs = radix * np.cos(angles) * np.sign(theta)
        ys = radix * np.sin(angles)

        current_angle = FlightPlanar.vec2rad(self.direction)
        path = np.c_[xs, ys]
        path = FlightPlanar.rot2d(Rotation.from_euler('z', current_angle - np.pi / 2), path)
        path += center
        path += self.position - path[0]

        self.add_path(path[:-1])
        self.position = path[-1]
        self.direction = FlightPlanar.rad2vec(current_angle + theta)

        if return_radix:
            return radix

    def cut_last_path(self, *, by_distance=None, by_steps=None):
        """切割上一条航段为两个航段

        Parameters
        ----------
        by_distance
            按指定长度序列依次切分航段，每个航段长度等同于数组每个元素
        by_steps
            按时间步数分割为若干航段，每个航段的时间步数等同于数组每个元素

        Notes
        -----
            若给定长度不够切分完整航段，则多余的部分单独成为一个航段；
            若总长度超出航段长度，则切完即停止，后续指定的长度会忽略，切割出的最后一个航段长度可能会小于指定长度；
            航段长度尽可能接近指定长度，但总是小于等于指定长度；
            如果指定单独的值，则视为长度为1的数组。
        """
        path = self.paths.pop(-1)
        if by_distance is not None:
            if not isinstance(by_distance, Iterable):
                by_distance = (by_distance,)

            by_steps = []
            for dist in by_distance:
                by_steps.append(dist // self.sample_distance)

        if not isinstance(by_steps, Iterable):
            by_steps = (by_steps,)

        idx = 0
        for n_steps in by_steps:
            end_idx = idx + n_steps
            self.paths.extend(path[idx:end_idx])
            idx = end_idx

    def split_last_path(self, *, by_distance=None, by_steps=None):
        """切割上一条航段

        :param by_distance: 按长度分割为若干航段
        :param by_steps: 按时间步数分割为若干航段
        """
        path = self.paths.pop(-1)
        if by_distance is not None:
            n_segments = len(path) * self.sample_distance // by_distance
        else:
            n_segments = len(path) // by_steps
        self.paths.extend(np.array_split(path, n_segments))

    def build(self):
        return np.concatenate(self.paths)

    def plot2d(self,
               screenshot: bool | PathLike = False,
               off_screen=False,
               arrows: bool | int = True,
               color: bool | 'Colormap' = False,
               alpha: float | Sequence[float] = 1,
               ax=None
               ):
        import matplotlib.pyplot as plt

        route = self.build()
        min_xy = np.min(route, axis=0)
        max_xy = np.max(route, axis=0)
        extent = max_xy - min_xy
        half = extent.max() / 2 * 1.1
        center = min_xy + extent / 2

        # TODO: change fig_size according to extent
        with check_axis(ax, show=False, figure=True, figsize=(5, 5)) as (fig, ax):
            if not color:
                ax.plot(*route.T, zorder=0)
            else:
                if color is True:
                    cmap = plt.get_cmap('tab10')
                else:
                    cmap = color

                if not isinstance(alpha, Sequence):
                    alpha = [alpha]

                n_colors = len(cmap.colors)
                n_alphas = len(alpha)
                for i, path_i in enumerate(self.paths):
                    ax.plot(
                        *path_i.T,
                        zorder=0,
                        color=cmap(i % n_colors),
                        alpha=alpha[i % n_alphas]
                    )

            if arrows:
                n_pts = route.shape[0]
                if arrows is True:
                    arrows = min(max(n_pts - 3, 1), 5)
                plot_arrows(ax, route, n_arrows=arrows, width=half * 0.01, zorder=1)

            ax.set_xlim(center[0] - half, center[0] + half)
            ax.set_ylim(center[1] - half, center[1] + half)

            if fig:
                if screenshot:
                    if screenshot is True:
                        import time
                        screenshot = Path(f'./{time.time_ns()}.png').absolute()
                    else:
                        screenshot = Path(screenshot).absolute()
                    fig.savefig(screenshot, bbox_inches='tight')

                if not off_screen:
                    fig.show()

                plt.close(fig)

    def plot2d_segments(self, path, **kwargs):
        for _ in self.plot2d_segments_customized(path=path, **kwargs):
            pass

    def plot2d_segments_customized(
            self,
            path,
            *,
            cmap=None,
            inactive_color=(0.7, 0.7, 0.7),
            inactive_alpha=0.3,
            show_index=True,
            fig=None
    ):
        from PIL import Image
        import tqdm
        import matplotlib.pyplot as plt
        from matplotlib.colors import ListedColormap

        if cmap is None:
            cmap = plt.get_cmap('tab10')

        if fig is None:
            fig, ax = plt.subplots(figsize=(5, 5))
        else:
            ax = fig.subplots()

        n_paths = len(self.paths)
        n_ratio = int(np.ceil(n_paths / len(cmap.colors)))
        colors = cmap.colors * n_ratio
        cmap_colors = [inactive_color] * n_paths
        alphas = [inactive_alpha] * n_paths

        pics = []
        for i in tqdm.trange(n_paths):
            yield ax  # 用于配置axis
            cmap_colors[i] = colors[i]
            alphas[i] = 1
            self.plot2d(ax=ax, color=ListedColormap(cmap_colors), alpha=alphas)

            if show_index:
                ax.text(*Coordinates(self.paths[i]).bounds.center, str(i), fontsize=10)

            fig.canvas.draw()
            size = fig.canvas.get_width_height() * np.asarray(fig.get_dpi()) / 100
            pics.append(Image.frombuffer("RGB", tuple(size.astype(int)), fig.canvas.tostring_rgb()))
            ax.cla()
            cmap_colors[i] = inactive_color
            alphas[i] = inactive_alpha

        pics[0].save(
            path,
            save_all=True,
            append_images=pics[1:],
            duration=500,
            loop=0
        )

    def plot2d_traversal(
        self,
        path,
        *,
        fig=None,
        frame_rate=24,
        total_time: int | float = 2,
    ):
        from PIL import Image
        from matplotlib import pyplot as plt
        import tqdm

        if fig is None:
            fig, ax = plt.subplots(figsize=(5, 5))
        else:
            ax = fig.add_subplot()

        route = self.build()
        min_xy = np.min(route, axis=0)
        max_xy = np.max(route, axis=0)
        extent = max_xy - min_xy
        half = extent.max() / 2 * 1.1
        center = min_xy + extent / 2

        ax.set_xlim(center[0] - half, center[0] + half)
        ax.set_ylim(center[1] - half, center[1] + half)

        ax.plot(*route.T, zorder=1)

        n_time_steps = len(route)
        n_frames = int(total_time * frame_rate)
        group_size = int(np.ceil(n_time_steps / n_frames))
        head_length = int(n_time_steps * 0.1)  # 用于限定着色位置为航线头部，避免航线较长时颜色混乱

        pics = []
        for i in tqdm.trange(0, n_time_steps, group_size, desc='Creating animation...'):
            rows = route[:i+group_size]
            colors = np.arange(-len(rows), 0)
            colors[:-head_length] = -head_length
            pr = ax.scatter(*rows.T, c=colors)
            fig.canvas.draw()
            size = fig.canvas.get_width_height() * np.asarray(fig.get_dpi()) / 100
            pics.append(Image.frombuffer("RGB", tuple(size.astype(int)), fig.canvas.tostring_rgb()))

            pr.remove()

        pics.extend([pics[-1]] * 5)  # 在结果上停留

        pics[0].save(
            path,
            save_all=True,
            append_images=pics[1:],
            duration=1000 / frame_rate,
            loop=0
        )

    @staticmethod
    def get_rot(theta):
        if theta >= 0:
            return FlightPlanar.Left
        else:
            return FlightPlanar.Right

    @staticmethod
    def rot2d(rot, vec2d, inverse=False):
        if vec2d.ndim == 1:
            return rot.apply(np.r_[vec2d, 0], inverse=inverse)[:2]
        else:
            return rot.apply(np.c_[vec2d, np.zeros(len(vec2d))], inverse=inverse)[:, :2]

    @staticmethod
    def vec2rad(vec):
        return math.atan2(vec[1], vec[0])

    @staticmethod
    def rad2vec(rad):
        return np.asarray((np.cos(rad), np.sin(rad)))

    @staticmethod
    def _spiral_arc_length(r, k):
        """等距螺旋线弧长不定积分

        .. math::
            \\int (\\sqrt{r(\\theta)^2 + r'(\\theta)^2}) d(\\theta)

            = \\int (\\sqrt{r^2 + k^2} / k) dr

            = (r \\sqrt{k^2 + r^2} + k^2 \\log(\\sqrt{k^2 + r^2} + r)/(2k) + constant
        """
        k2 = k ** 2
        p = np.sqrt(k2 + r ** 2)
        return (r * p + k2 * np.log(p + r)) / (2 * k)

    @staticmethod
    def _spiral_arc_length_derive(r, k):
        """等距螺旋线弧长不定积分的导数，即待积式

        .. math::
            \\sqrt{r^2 + k^2} / k
        """
        return np.sqrt(k ** 2 + r ** 2) / k


def plot_arrows(ax, path, n_arrows=None, width=10, cmap=None, **kwargs):
    n_pts = path.shape[0]
    if n_arrows is None:
        n_arrows = min(max(n_pts - 3, 1), 5)
    arrow_locs = np.linspace(n_pts * 0.02, n_pts * 0.98, n_arrows).astype(int)

    if cmap is None:
        from matplotlib.colors import ListedColormap
        cmap = ListedColormap(['red'] + ['black'] * (n_arrows - 1))
    else:
        if cmap.N != n_arrows:
            cmap = cmap._resample(n_arrows)

    first_color = 'red'
    normal_color = 'black'
    color = first_color

    for i, idx in enumerate(arrow_locs):
        start = path[idx]
        end = path[idx + 1]
        ax.arrow(*start, *(end - start),
                 width=width,
                 facecolor=cmap(i),
                 edgecolor=cmap(i),
                 **kwargs)

        if color == first_color:
            color = normal_color



from . import Shape3D, PlateFramework, Composition, BarFramework


class PillarFramework(PlateFramework):
    def __init__(self,
                 outline: Shape3D,
                 spec,
                 n_rooms=None,
                 room_spec=None,
                 bounds=None,
                 *,
                 inherit_transform=False
                 ):
        """定义一个楼板-框架结构，其边界由 `outline` 决定，并基于 `outline` 裁剪边界。
        """
        super().__init__(
            outline, spec, n_rooms, room_spec, bounds,
            inherit_transform=inherit_transform
        )

    def _create_base_framework(self):
        return Composition(
            *PlateFramework._create_parts(self, 2),
            *BarFramework._create_parts(self, 2),
            mix_mode=Composition.Union
        )



import numpy as np
from PIL import Image
from pyproj import CRS

from metalpy.utils.bounds import Bounds


class GeoImageRefSystem:
    def __init__(self, origin, unit_size, crs=None, flip_y=None):
        """用于表示地理图像的各个像素中心点和地理坐标的映射关系

        Parameters
        ----------
        origin
            地理图像参考系的原点，以像素边缘为边界
        unit_size
            地理图像参考系每个单元（像素）对应的地理距离
        crs
            地理参考坐标系
        flip_y
            指示地理坐标系与图像坐标系的y轴方向是否相反

        Notes
        -----
        地理坐标系保证与单位网格坐标系方向相同，图像坐标系只在涉及图像处理时才会引入，此时才会考虑`flip_y`等问题
        """
        unit_size = np.array(unit_size)
        origin = np.array(origin)

        assert unit_size[0] > 0, 'Flipping of X axis is not allowed.'
        if flip_y is None:
            flip_y = unit_size[1] < 0

        self.origin = origin
        self.unit_size = np.abs(unit_size)
        self.flip_y = flip_y

        if crs is not None:
            if isinstance(crs, str):
                crs = CRS.from_string(crs)
            elif isinstance(crs, int):
                crs = CRS.from_epsg(crs)
            self.crs = crs
        else:
            self.crs = CRS.from_string('WGS 84')

    def create_image(self, image_size=None, offset=(0, 0), bounds=None, geo_bounds=None):
        from metalpy.carto.basemap.geo_image import GeoImage

        if bounds is not None:
            bounds = Bounds(bounds)
            offset = bounds.origin
            image_size = bounds.extent
        elif geo_bounds is not None:
            bounds = self.to_unit_bounds(geo_bounds)
            offset = bounds.origin
            image_size = bounds.extent
        else:
            image_size = np.asarray(image_size)
            offset = np.asarray(offset)

        image_size = image_size.astype(int)
        offset = offset.astype(int)

        return GeoImage(Image.new('RGBA', tuple(image_size)), offset, self)

    def map_image(self, image: Image.Image, offset=(0, 0), geo_offset=None):
        from metalpy.carto.basemap.geo_image import GeoImage

        if geo_offset is not None:
            offset = self.to_unit_bounds(np.c_[geo_offset, (0, 0)].ravel())

        return GeoImage(image, offset, self)

    def to_unit_bounds(self, bounds):
        corners = (Bounds(bounds).as_corners() - self.origin) / self.unit_size
        corners.origin = np.floor(corners.origin)
        corners.end = np.ceil(corners.end)
        return corners.as_bounds()

    def to_geo_bounds(self, bounds):
        corners = Bounds(bounds).as_corners() * self.unit_size + self.origin
        return corners.as_bounds()

    def to_edge_geo_bounds(self, geo_bounds):
        return self.to_geo_bounds(self.to_unit_bounds(geo_bounds))

    @staticmethod
    def of_unit_edge_bounds(edge_origin, unit_size, crs=None, flip_y=None) -> 'GeoImageRefSystem':
        """
        Notes
        -----
                                 ↓ end[right top]
            ┌────┬────┬─────┬────┐ ←
            │    │    │ ... │    │
            ├────┼────┼─────┼────┤
            │    │    │ ... │    │
            ├────┼────┼─────┼────┤
            │    │    │ ... │    │
          → └────┴────┴─────┴────┘
            ↑ origin[left bottom]
        """
        return GeoImageRefSystem(edge_origin, unit_size, crs=crs, flip_y=flip_y)

    @staticmethod
    def of_unit_center_bounds(center_origin, unit_size, crs=None, flip_y=None) -> 'GeoImageRefSystem':
        """
        Notes
        -----
                              ↓ end[right top]
            ┌────┬────┬─────┬────┐
            │    │    │ ... │    │ ←
            ├────┼────┼─────┼────┤
            │    │    │ ... │    │
            ├────┼────┼─────┼────┤
          → │    │    │ ... │    │
            └────┴────┴─────┴────┘
              ↑ origin[left bottom]
        """
        return GeoImageRefSystem(
            center_origin - np.asarray(unit_size) / 2,
            unit_size,
            crs=crs,
            flip_y=flip_y
        )



import numpy as np

from scipy.stats import linregress

from metalpy.utils.bounds import Bounds
from metalpy.utils.dhash import dhash
from metalpy.utils.ear_clip import ear_clip
from metalpy.utils.geometry import gen_random_convex_polygon
from . import Shape3D


def is_abs_distance_in(arr, x0, r):
    return np.abs(arr - x0) < r


def is_in_multipoints_or_point(pt, mp_or_pt):
    if mp_or_pt.geom_type == 'Point':
        return pt == mp_or_pt
    else:
        return pt in mp_or_pt.geoms


class Edge:
    def __init__(self, p1, p2):
        self.p1, self.p2 = p1, p2
        self.ymin = min(p1[1], p2[1])
        self.ymax = max(p1[1], p2[1])
        self.xmin = min(p1[0], p2[0])
        self.xmax = max(p1[0], p2[0])

        ends = np.vstack([p1, p2])

        if p1[0] != p2[0]:  # 非垂直于x轴
            slope, intercept, _, _, _ = linregress(ends[:, 0], ends[:, 1])
            self.k, self.b = slope, intercept
        else:
            self.k = None
            self.x0 = p1[0]

    def intersects(self, pts):
        x, y = pts[:, 0], pts[:, 1]
        if self.k is None:
            # 处理与垂直线相交的情形
            # 由于是多面体，因此与该直线相交的射线一定与其重合
            # 其所在直线会与该边的相邻两个边也存在交点
            # 交于下端点视为相交，其余位置不视为相交
            # 可以保证边缘点被视为多边形内
            # 1. 点p在该线段之下，则只会与相邻边有2个交点
            # 2. 点p在该线段与下方邻边交点，则会与相邻边和该边有3个交点
            # 3. 点p在该线段中间及上方邻边交点，则只会与上方邻边有1个交点
            indices = x == self.x0
            indices[indices] = y[indices] == self.ymin
            return indices
        elif self.k == 0:
            return (x >= self.xmin) & (x <= self.xmax) & (y <= self.ymin)
        else:
            y_intersection = self.k * x + self.b
            return (y_intersection >= self.ymin) & (y_intersection <= self.ymax) & (y <= y_intersection)


class Prism(Shape3D):
    def __init__(self, pts, z0, z1, verbose=True, cells=None):
        """定义任意底面棱锥

        Parameters
        ----------
        pts
            底面多边形顶点列表
        z0, z1
            底面和顶面高度
        verbose
            是否输出辅助信息（主要为顶点数较多时导出PyVista模型会触发进度条）
        cells
            array(n-2, 3)，每行包含三个属于(0, n-1)的下标，指定pts三角化后的结果
        """
        super().__init__()
        self.pts = np.asarray(pts)
        self.z0 = min(z0, z1)
        self.z1 = max(z0, z1)

        if cells is not None:
            cells_shape = (len(self.pts) - 2, 3)
            cells = np.asarray(cells)
            assert cells.shape == cells_shape, \
                f'`cells` must have shape {cells_shape},' \
                f' containing triangulated polygon cells. (got {cells.shape})'
            self.triangulated_polygon = np.asarray(cells)
        else:
            self.triangulated_polygon = ear_clip(self.pts, verbose=verbose)

    @staticmethod
    def rand(n_edges,
             size=(1, 1),
             z0=0,
             z1=1,
             random_state=None,
             ):
        """生成随机棱柱

        Parameters
        ----------
        n_edges
            随机棱柱的边数
        size
            随机棱柱在xy平面上的尺寸
        z0, z1
            随机棱柱的底面和顶面高度
        random_state
            随机种子

        Returns
        -------
        ret
            随机棱柱实例
        """
        pts = gen_random_convex_polygon(n_edges, size=size, random_state=random_state)
        return Prism(pts, z0, z1)

    @property
    def h(self):
        return self.z1 - self.z0

    def do_place(self, mesh_cell_centers, progress):
        n_possible_grids = len(mesh_cell_centers)

        # 优化: 使用射线法，取向+y方向的射线
        if n_possible_grids > 0:
            edges = []
            n_edges = len(self.pts)
            for i in range(n_edges):
                edges.append(Edge(self.pts[i], self.pts[(i + 1) % n_edges]))

            mesh = mesh_cell_centers[:, 0:2]
            n_intersects = np.zeros(n_possible_grids, dtype=int)

            for edge in edges:
                intercepts = edge.intersects(mesh)
                n_intersects = n_intersects + intercepts

            indices_horizontally_satisfied = (n_intersects & 1) == 1
        else:
            indices_horizontally_satisfied = np.full(n_possible_grids, False)

        return indices_horizontally_satisfied

    def __dhash__(self):
        return dhash(
            super().__dhash__(),
            self.z0, self.z1,
            *self.pts.ravel(),
            self.triangulated_polygon
        )

    def do_clone(self, deep=True):
        return Prism(self.pts.copy(), self.z0, self.z1)

    @property
    def local_bounds(self):
        return Bounds(np.c_[
            np.r_[self.pts.min(axis=0), self.z0],
            np.r_[self.pts.max(axis=0), self.z1]
        ].ravel())

    def to_local_polydata(self):
        import pyvista as pv

        pts = self.pts
        z0, z1 = self.z0, self.z1
        n_pts = len(pts)
        n_vertices = 2 * len(pts)
        vh = pts.repeat(2, axis=0)
        vz = np.asarray([[z0, z1]]).repeat(n_pts, axis=0).ravel()
        vertices = np.c_[vh, vz]
        indices = np.arange(0, n_vertices, 1).reshape([-1, 2]).T

        faces = self.triangulated_polygon
        top_face = np.c_[np.repeat(3, faces.shape[0]), faces * 2].ravel()  # 上顶面
        bottom_face = np.c_[np.repeat(3, faces.shape[0]), faces * 2 + 1].ravel()  # 下底面
        edge_counts = np.ones(n_pts, dtype=np.integer) * 4
        side_faces = np.c_[
            edge_counts,
            indices[0], indices[1],
            np.roll(indices[1], -1), np.roll(indices[0], -1)
        ].ravel()

        faces = np.r_[top_face, bottom_face, side_faces]

        shape = pv.PolyData(vertices.astype(np.float64), faces=faces)
        return shape

    @property
    def bottom_area(self):
        s = 0
        for vi in self.triangulated_polygon:
            v = self.pts[vi]
            si = np.abs(np.cross(v[1] - v[0], v[2] - v[0])) / 2
            s += si

        return s

    @property
    def volume(self):
        return self.bottom_area * self.h

    @property
    def area(self):
        ba = self.bottom_area
        rolled_pts = np.roll(self.pts, 1, axis=0)
        perimeter = np.linalg.norm(self.pts - rolled_pts, axis=1).sum()

        return 2 * ba + perimeter * self.h
